from __future__ import annotations

import datetime as dt
import hashlib
import random
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol, cast

import pyray as rl

from grim.assets import PaqTextureCache, TextureLoader
from grim.audio import AudioState, play_music
from grim.config import (
    CrimsonConfig,
)
from grim.console import ConsoleState
from grim.fonts.grim_mono import GrimMonoFont, load_grim_mono_font
from grim.geom import Rect, Vec2
from grim.math import clamp
from grim.view import ViewContext

from ..debug import debug_enabled
from ..game_modes import GameMode
from ..input_codes import (
    config_keybinds_for_player,
    input_code_is_down_for_player,
    input_code_is_pressed_for_player,
    input_primary_just_pressed,
)
from ..net.protocol import STATE_HASH_PERIOD_TICKS, TickFrame
from ..perks.state import CreatureForPerks
from ..persistence.save_status import GameStatus
from ..quests import quest_by_level
from ..quests.runtime import build_quest_spawn_table
from ..quests.runtime import tick_quest_completion_transition as _legacy_tick_quest_completion_transition
from ..quests.timeline import quest_spawn_table_empty as _legacy_quest_spawn_table_empty
from ..quests.timeline import tick_quest_mode_spawns as _legacy_tick_quest_mode_spawns
from ..quests.types import QuestContext, QuestDefinition, SpawnEntry
from ..replay import ReplayHeader, ReplayRecorder, ReplayStatusSnapshot, dump_replay
from ..replay.checkpoints import (
    FORMAT_VERSION as CHECKPOINTS_FORMAT_VERSION,
)
from ..replay.checkpoints import (
    ReplayCheckpoint,
    ReplayCheckpoints,
    build_checkpoint,
    default_checkpoints_path,
    dump_checkpoints_file,
    resolve_checkpoint_sample_rate,
)
from ..replay.input_codec import pack_player_input, unpack_player_input
from ..replay.types import WEAPON_USAGE_COUNT
from ..sim.clock import FixedStepClock
from ..sim.input import PlayerInput
from ..sim.sessions import QuestDeterministicSession, QuestDeterministicSessionTick
from ..terrain_assets import TerrainTextureId, terrain_texture_by_id
from ..ui.cursor import draw_menu_cursor
from ..ui.hud import draw_hud_overlay, hud_flags_for_game_mode
from ..ui.perk_menu import PerkMenuAssets, load_perk_menu_assets
from ..views.quest_title_overlay import draw_quest_title_overlay
from ..weapon_runtime import (
    most_used_weapon_id_for_player,
    weapon_assign_player,
)
from ..weapons import WEAPON_BY_ID
from .base_gameplay_mode import BaseGameplayMode
from .components.highscore_record_builder import shots_from_state
from .components.perk_menu_controller import PerkMenuContext, PerkMenuController
from .components.perk_prompt_ui import PERK_PROMPT_MAX_TIMER_MS, PerkPromptUi

WORLD_SIZE = 1024.0
QUEST_TITLE_FADE_IN_MS = 500.0
QUEST_TITLE_HOLD_MS = 1000.0
QUEST_TITLE_FADE_OUT_MS = 500.0
QUEST_TITLE_TOTAL_MS = QUEST_TITLE_FADE_IN_MS + QUEST_TITLE_HOLD_MS + QUEST_TITLE_FADE_OUT_MS

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_SPONSOR_COLOR = rl.Color(255, 255, 255, int(255 * 0.5))

_DEBUG_WEAPON_IDS = tuple(sorted(WEAPON_BY_ID))
QUEST_COMPLETE_BANNER_BASE_W = 256.0
QUEST_COMPLETE_BANNER_BASE_H = 32.0
QUEST_COMPLETE_BANNER_SCALE_BASE = 0.95
QUEST_COMPLETE_BANNER_SCALE_RATE = 0.0004 * 0.13
QUEST_COMPLETE_BANNER_FADE_IN_MS = 500.0
QUEST_COMPLETE_BANNER_HOLD_END_MS = 1500.0
QUEST_COMPLETE_BANNER_FADE_OUT_END_MS = 2000.0

# Compatibility aliases used by existing monkeypatch-based tests.
tick_quest_mode_spawns = _legacy_tick_quest_mode_spawns
tick_quest_completion_transition = _legacy_tick_quest_completion_transition
quest_spawn_table_empty = _legacy_quest_spawn_table_empty


class QuestSessionLike(Protocol):
    detail_preset: int
    fx_toggle: int
    spawn_entries: tuple[SpawnEntry, ...]
    spawn_timeline_ms: float
    no_creatures_timer_ms: float
    completion_transition_ms: float

    def step_tick(
        self,
        *,
        dt_frame: float,
        inputs: list[PlayerInput] | None,
        trace_rng: bool = False,
    ) -> QuestDeterministicSessionTick: ...


@dataclass(slots=True)
class _QuestRunState:
    quest: QuestDefinition | None = None
    level: str = ""
    spawn_entries: tuple[SpawnEntry, ...] = ()
    total_spawn_count: int = 0
    max_trigger_time_ms: int = 0
    spawn_timeline_ms: float = 0.0
    quest_name_timer_ms: float = 0.0
    no_creatures_timer_ms: float = 0.0
    completion_transition_ms: float = -1.0


@dataclass(frozen=True, slots=True)
class QuestRunOutcome:
    kind: str  # "completed" | "failed"
    level: str
    base_time_ms: int
    player_health: float
    player2_health: float | None
    pending_perk_count: int
    experience: int
    kill_count: int
    weapon_id: int
    shots_fired: int
    shots_hit: int
    most_used_weapon_id: int
    player_health_values: tuple[float, ...] = ()


def _quest_seed(major: int, minor: int) -> int:
    return int(major) * 100 + int(minor)


def _quest_attempt_counter_index(major: int, minor: int) -> int | None:
    tier = int(major)
    quest = int(minor)
    global_index = (tier - 1) * 10 + (quest - 1)
    if not (0 <= global_index < 40):
        return None
    return global_index + 11


def _quest_level_label(major: int, minor: int) -> str:
    major = int(major)
    minor = int(minor)

    # Match `ui_render_hud` (0x0041bf94): quest minor can temporarily exceed 10
    # (e.g. after incrementing), and the HUD carries it into the major.
    while minor > 10:
        major += 1
        minor -= 10
    return f"{major}.{minor}"


def _quest_complete_banner_alpha(timer_ms: float) -> float:
    t = float(timer_ms)
    if t <= 0.0:
        return 0.0
    if t < QUEST_COMPLETE_BANNER_FADE_IN_MS:
        return clamp(t / QUEST_COMPLETE_BANNER_FADE_IN_MS, 0.0, 1.0)
    if t < QUEST_COMPLETE_BANNER_HOLD_END_MS:
        return 1.0
    if t < QUEST_COMPLETE_BANNER_FADE_OUT_END_MS:
        return clamp((QUEST_COMPLETE_BANNER_FADE_OUT_END_MS - t) / QUEST_COMPLETE_BANNER_FADE_IN_MS, 0.0, 1.0)
    return 0.0


class QuestMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        demo_mode_active: bool = False,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=int(GameMode.QUESTS),
            demo_mode_active=bool(demo_mode_active),
            difficulty_level=0,
            hardcore=False,
            texture_cache=texture_cache,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._quest = _QuestRunState()
        self._selected_level: str | None = None
        self._outcome: QuestRunOutcome | None = None
        self._perk_menu_assets: PerkMenuAssets | None = None
        self._grim_mono: GrimMonoFont | None = None
        self._quest_complete_texture: rl.Texture | None = None

        self._perk_prompt_timer_ms = 0.0
        self._perk_prompt_hover = False
        self._perk_prompt_pulse = 0.0
        self._perk_menu = PerkMenuController(on_close=self._reset_perk_prompt, on_pick=self._record_perk_pick)
        self._sim_clock = FixedStepClock(tick_rate=60)
        self._lan_capture_clock = FixedStepClock(tick_rate=60)
        self._sim_session: QuestSessionLike | None = None
        self._replay_recorder: ReplayRecorder | None = None
        self._replay_checkpoints: list[ReplayCheckpoint] = []
        self._replay_checkpoints_sample_rate: int = 60
        self._replay_checkpoints_last_tick: int | None = None

    def open(self) -> None:
        super().open()
        self._quest = _QuestRunState()
        self._outcome = None
        self._perk_menu_assets = load_perk_menu_assets(self._assets_root)
        self._quest_complete_texture = self._load_quest_complete_texture()
        self._grim_mono = load_grim_mono_font(self._assets_root)

        self._perk_prompt_timer_ms = 0.0
        self._perk_prompt_hover = False
        self._perk_prompt_pulse = 0.0
        self._perk_menu.reset()
        self._sim_clock.reset()
        self._lan_capture_clock.reset()
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        self._sim_session = QuestDeterministicSession(
            world=self.world.world_state,
            world_size=float(self.world.world_size),
            damage_scale_by_type=self.world._damage_scale_by_type,
            fx_queue=self.world.fx_queue,
            fx_queue_rotated=self.world.fx_queue_rotated,
            spawn_entries=(),
            detail_preset=5,
            fx_toggle=0,
            clear_fx_queues_each_tick=False,
        )

    def close(self) -> None:
        if self._grim_mono is not None:
            rl.unload_texture(self._grim_mono.texture)
            self._grim_mono = None
        self._quest_complete_texture = None
        self._perk_menu_assets = None
        self._sim_session = None
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        super().close()

    def _load_quest_complete_texture(self) -> rl.Texture | None:
        loader = TextureLoader(
            assets_root=self._assets_root,
            cache=self.world.texture_cache,
        )
        texture = loader.get(
            name="ui_textLevComp",
            paq_rel="ui/ui_textLevComp.jaz",
        )
        return texture

    def _reset_perk_prompt(self) -> None:
        if int(self.state.perk_selection.pending_count) > 0:
            # Reset the prompt swing so each pending perk replays the intro.
            self._perk_prompt_timer_ms = 0.0
            self._perk_prompt_hover = False
            self._perk_prompt_pulse = 0.0

    def _record_perk_pick(self, choice_index: int) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        recorder.record_perk_pick(player_index=0, choice_index=int(choice_index))

    def _record_replay_checkpoint(
        self,
        tick_index: int,
        *,
        force: bool = False,
        rng_marks: dict[str, int] | None = None,
        deaths: Sequence[object] | None = None,
        events: object | None = None,
        command_hash: str | None = None,
    ) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        if tick_index < 0:
            return
        if not force and (tick_index % int(self._replay_checkpoints_sample_rate or 1)) != 0:
            return
        if self._replay_checkpoints_last_tick == int(tick_index):
            return
        self._replay_checkpoints.append(
            build_checkpoint(
                tick_index=int(tick_index),
                world=self.world.world_state,
                elapsed_ms=float(self._quest.spawn_timeline_ms),
                rng_marks=rng_marks,
                deaths=deaths,
                events=events,
                command_hash=command_hash,
            ),
        )
        self._replay_checkpoints_last_tick = int(tick_index)

    def _save_replay(self) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        if int(recorder.tick_index) <= 0:
            # Avoid emitting empty replays/checkpoint sidecars (usually indicates a
            # test harness calling failure/complete helpers without ticking).
            self._replay_recorder = None
            self._replay_checkpoints.clear()
            self._replay_checkpoints_last_tick = None
            return
        replay = recorder.finish()
        self._record_replay_checkpoint(max(0, recorder.tick_index - 1), force=True)
        terminal_tick = int(recorder.tick_index)
        if any(int(event.tick_index) == terminal_tick for event in replay.events):
            self._record_replay_checkpoint(terminal_tick, force=True)
        data = dump_replay(replay)
        digest = hashlib.sha256(data).hexdigest()
        stamp = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
        replay_dir = self._base_dir / "replays"
        replay_dir.mkdir(parents=True, exist_ok=True)
        level = str(self._quest.level) if self._quest.level else str(replay.header.quest_level or "quest")
        kind = str(self._outcome.kind) if self._outcome is not None else "quest"
        base_time_ms = int(self._quest.spawn_timeline_ms)
        base_name = f"quest_{level}_{stamp}_{kind}_t{base_time_ms}"
        path = replay_dir / f"{base_name}.crd"
        counter = 1
        while path.exists():
            path = replay_dir / f"{base_name}_{counter}.crd"
            counter += 1
        path.write_bytes(data)
        checkpoints_path = default_checkpoints_path(path)
        dump_checkpoints_file(
            checkpoints_path,
            ReplayCheckpoints(
                version=CHECKPOINTS_FORMAT_VERSION,
                replay_sha256=digest,
                sample_rate=int(self._replay_checkpoints_sample_rate or 0),
                checkpoints=list(self._replay_checkpoints),
            ),
        )
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        if self._console is not None:
            self._console.log.log(f"replay: saved {path}")
            self._console.log.log(f"replay: saved {checkpoints_path}")
            self._console.log.flush()

    def _perk_menu_context(self) -> PerkMenuContext:
        fx_toggle = self.config.fx_toggle
        fx_detail = self.config.fx_detail(level=0, default=False)
        players = self.world.players
        return PerkMenuContext(
            state=self.state,
            perk_state=self.state.perk_selection,
            players=players,
            creatures=cast("list[CreatureForPerks]", self.creatures.entries),
            player=self.player,
            game_mode=int(GameMode.QUESTS),
            player_count=len(players),
            fx_toggle=fx_toggle,
            fx_detail=fx_detail,
            font=self._small,
            assets=self._perk_menu_assets,
            mouse=self._ui_mouse_pos(),
            play_sfx=self.world.audio_router.play_sfx,
        )

    def select_level(self, level: str | None) -> None:
        self._selected_level = level

    def consume_outcome(self) -> QuestRunOutcome | None:
        outcome = self._outcome
        self._outcome = None
        return outcome

    def prepare_new_run(self, level: str, *, status: GameStatus | None) -> None:
        quest = quest_by_level(level)
        if quest is None:
            self._quest = _QuestRunState(level=level)
            return
        self._outcome = None
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

        hardcore_flag = self.config.hardcore

        self.world.hardcore = hardcore_flag
        seed = _quest_seed(quest.major, quest.minor)

        player_count = self.config.player_count
        self.world.reset(seed=seed, player_count=max(1, min(4, player_count)))
        self._bind_world()
        self._local_input.reset(players=self.world.players)
        self.bind_status(status)
        self.state.quest_stage_major, self.state.quest_stage_minor = quest.level_key

        base_id, overlay_id, detail_id = quest.terrain_ids or (
            TerrainTextureId.Q1_BASE,
            TerrainTextureId.Q1_OVERLAY,
            TerrainTextureId.Q1_BASE,
        )
        base = terrain_texture_by_id(int(base_id))
        overlay = terrain_texture_by_id(int(overlay_id))
        detail = terrain_texture_by_id(int(detail_id))
        if base is not None and overlay is not None:
            base_key, base_path = base
            overlay_key, overlay_path = overlay
            detail_key = detail[0] if detail is not None else None
            detail_path = detail[1] if detail is not None else None
            self.world.set_terrain(
                base_key=base_key,
                overlay_key=overlay_key,
                base_path=base_path,
                overlay_path=overlay_path,
                detail_key=detail_key,
                detail_path=detail_path,
            )

        # Quest metadata already stores native (1-based) weapon ids.
        start_weapon_id = max(1, int(quest.start_weapon_id))
        for player in self.world.players:
            weapon_assign_player(player, start_weapon_id)

        ctx = QuestContext(
            width=int(self.world.world_size),
            height=int(self.world.world_size),
            player_count=len(self.world.players),
        )
        entries = build_quest_spawn_table(
            quest,
            ctx,
            seed=seed,
            hardcore=hardcore_flag,
            full_version=not self.world.demo_mode_active,
        )
        total_spawn_count = sum(int(entry.count) for entry in entries)
        max_trigger_ms = max((int(entry.trigger_ms) for entry in entries), default=0)

        self._quest = _QuestRunState(
            quest=quest,
            level=quest.level,
            spawn_entries=entries,
            total_spawn_count=int(total_spawn_count),
            max_trigger_time_ms=int(max_trigger_ms),
            spawn_timeline_ms=0.0,
            quest_name_timer_ms=0.0,
            no_creatures_timer_ms=0.0,
            completion_transition_ms=-1.0,
        )
        self._sim_clock.reset()
        self._sim_session = QuestDeterministicSession(
            world=self.world.world_state,
            world_size=float(self.world.world_size),
            damage_scale_by_type=self.world._damage_scale_by_type,
            fx_queue=self.world.fx_queue,
            fx_queue_rotated=self.world.fx_queue_rotated,
            spawn_entries=tuple(entries),
            detail_preset=5,
            fx_toggle=0,
            clear_fx_queues_each_tick=False,
        )

        weapon_usage_counts: tuple[int, ...] = ()
        if status is not None:
            raw_counts = status.data.get("weapon_usage_counts")
            if isinstance(raw_counts, list):
                coerced: list[int] = []
                for value in raw_counts[:WEAPON_USAGE_COUNT]:
                    try:
                        coerced.append(int(value) & 0xFFFFFFFF)
                    except (TypeError, ValueError, OverflowError):
                        coerced.append(0)
                weapon_usage_counts = tuple(coerced)
        if len(weapon_usage_counts) != WEAPON_USAGE_COUNT:
            weapon_usage_counts = tuple(weapon_usage_counts) + (0,) * max(
                0, WEAPON_USAGE_COUNT - len(weapon_usage_counts),
            )
            weapon_usage_counts = weapon_usage_counts[:WEAPON_USAGE_COUNT]
        status_snapshot = ReplayStatusSnapshot(
            quest_unlock_index=int(status.quest_unlock_index) if status is not None else 0,
            quest_unlock_index_full=int(status.quest_unlock_index_full)
            if status is not None
            else 0,
            weapon_usage_counts=weapon_usage_counts,
        )
        record_replay = (not bool(self._lan_enabled)) or str(self._lan_role) == "host"
        if record_replay:
            self._replay_recorder = ReplayRecorder(
                ReplayHeader(
                    game_mode_id=int(GameMode.QUESTS),
                    seed=int(self.state.rng.state),
                    quest_level=str(quest.level),
                    tick_rate=int(self._sim_clock.tick_rate),
                    difficulty_level=int(self.world.difficulty_level),
                    hardcore=bool(self.world.hardcore),
                    preserve_bugs=bool(self.world.preserve_bugs),
                    detail_preset=self.config.detail_preset,
                    fx_toggle=self.config.fx_toggle,
                    world_size=float(self.world.world_size),
                    player_count=len(self.world.players),
                    status=status_snapshot,
                ),
            )
            tick_rate = int(self._replay_recorder.header.tick_rate)
            self._replay_checkpoints_sample_rate = resolve_checkpoint_sample_rate(tick_rate)
        else:
            self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

        if status is not None:
            idx = _quest_attempt_counter_index(quest.major, quest.minor)
            if idx is not None:
                status.increment_quest_play_count(idx)

    def _handle_input(self) -> None:
        if self._perk_menu.open and rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self.world.audio_router.play_sfx("sfx_ui_buttonclick")
            self._perk_menu.close()
            return

        if (not bool(self._lan_enabled)) and rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if debug_enabled() and (not self._perk_menu.open):
            if rl.is_key_pressed(rl.KeyboardKey.KEY_F2):
                self.state.debug_god_mode = not bool(self.state.debug_god_mode)
                self.world.audio_router.play_sfx("sfx_ui_buttonclick")
            if rl.is_key_pressed(rl.KeyboardKey.KEY_F3):
                self.state.perk_selection.pending_count += 1
                self.state.perk_selection.choices_dirty = True
                self.world.audio_router.play_sfx("sfx_ui_levelup")
            if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT_BRACKET):
                self._debug_cycle_weapon(-1)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT_BRACKET):
                self._debug_cycle_weapon(1)

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _debug_cycle_weapon(self, delta: int) -> None:
        weapon_ids = _DEBUG_WEAPON_IDS
        if not weapon_ids:
            return
        current = int(self.player.weapon_id)
        try:
            idx = weapon_ids.index(current)
        except ValueError:
            idx = 0
        weapon_id = int(weapon_ids[(idx + int(delta)) % len(weapon_ids)])
        weapon_assign_player(self.player, weapon_id, state=self.state)

    def _death_transition_ready(self) -> bool:
        dead_players = 0
        for player in self.world.players:
            if float(player.health) > 0.0:
                return False
            dead_players += 1
            if float(player.death_timer) >= 0.0:
                return False
        return dead_players > 0

    def _tick_death_timers(self, dt_frame: float, *, rate: float = 20.0) -> None:
        delta = float(dt_frame) * float(rate)
        if delta <= 0.0:
            return
        for player in self.world.players:
            if float(player.health) > 0.0:
                continue
            if float(player.death_timer) < 0.0:
                continue
            player.death_timer = float(player.death_timer) - delta

    def _close_failed_run(self) -> None:
        if self._outcome is None:
            fired, hit = shots_from_state(self.state, player_index=int(self.player.index))
            most_used_weapon_id = most_used_weapon_id_for_player(
                self.state,
                player_index=int(self.player.index),
                fallback_weapon_id=int(self.player.weapon_id),
            )
            player_health_values = tuple(float(player.health) for player in self.world.players)
            player2_health = None
            if len(player_health_values) >= 2:
                player2_health = float(player_health_values[1])
            self._outcome = QuestRunOutcome(
                kind="failed",
                level=str(self._quest.level),
                base_time_ms=int(self._quest.spawn_timeline_ms),
                player_health=float(player_health_values[0] if player_health_values else self.player.health),
                player2_health=player2_health,
                player_health_values=player_health_values,
                pending_perk_count=int(self.state.perk_selection.pending_count),
                experience=int(self.player.experience),
                kill_count=int(self.creatures.kill_count),
                weapon_id=int(self.player.weapon_id),
                shots_fired=fired,
                shots_hit=hit,
                most_used_weapon_id=int(most_used_weapon_id),
            )
        self._save_replay()
        self.close_requested = True

    def _draw_perk_prompt(self) -> None:
        if self._perk_menu.active:
            return
        if not any(player.health > 0.0 for player in self.world.players):
            return
        pending_count = int(self.state.perk_selection.pending_count)
        if pending_count <= 0:
            return
        label = PerkPromptUi.label(self.config, pending_count=pending_count)
        if not label:
            return
        PerkPromptUi.draw(
            font=self._small,
            assets=self._perk_menu_assets,
            label=label,
            timer_ms=float(self._perk_prompt_timer_ms),
            pulse=float(self._perk_prompt_pulse),
            ui_text_width=self._ui_text_width,
            text_color=UI_TEXT_COLOR,
            scale=UI_TEXT_SCALE,
        )

    def update(self, dt: float) -> None:
        self._update_audio(dt)

        dt_frame, dt_ui_ms = self._tick_frame(dt)
        self._handle_input()
        if self._action == "open_pause_menu":
            return

        if self.close_requested:
            return

        if bool(self._lan_enabled) and self._lan_runtime is not None:
            self._update_lan_match(dt_frame=dt_frame, dt_ui_ms=dt_ui_ms)
            return

        any_alive = any(player.health > 0.0 for player in self.world.players)
        perk_pending = int(self.state.perk_selection.pending_count) > 0 and any_alive

        self._perk_prompt_hover = False
        perk_ctx = self._perk_menu_context()
        if self._perk_menu.open:
            self._perk_menu.handle_input(perk_ctx, dt_frame=dt_frame, dt_ui_ms=dt_ui_ms)

        perk_menu_active = self._perk_menu.active

        if (not perk_menu_active) and perk_pending and (not self._paused):
            label = PerkPromptUi.label(self.config, pending_count=int(self.state.perk_selection.pending_count))
            if label:
                rect = PerkPromptUi.rect(
                    label,
                    ui_text_width=self._ui_text_width,
                    ui_line_height=self._ui_line_height,
                    assets=self._perk_menu_assets,
                    scale=UI_TEXT_SCALE,
                )
                self._perk_prompt_hover = rect.contains(self._ui_mouse_pos())

            player0_binds = config_keybinds_for_player(self.config, player_index=0)
            fire_key = 0x100
            if len(player0_binds) >= 5:
                fire_key = int(player0_binds[4])

            pick_key = self.config.keybind_pick_perk

            if input_code_is_pressed_for_player(pick_key, player_index=0) and (
                not input_code_is_down_for_player(fire_key, player_index=0)
            ):
                self._perk_prompt_pulse = 1000.0
                if self._replay_recorder is not None:
                    self._record_replay_checkpoint(max(0, self._replay_recorder.tick_index - 1), force=True)
                opened = self._perk_menu.open_if_available(perk_ctx)
                if opened and self._replay_recorder is not None:
                    self._replay_recorder.record_perk_menu_open(player_index=0)
            elif self._perk_prompt_hover and input_primary_just_pressed(
                self.config,
                player_count=len(self.world.players),
            ):
                self._perk_prompt_pulse = 1000.0
                if self._replay_recorder is not None:
                    self._record_replay_checkpoint(max(0, self._replay_recorder.tick_index - 1), force=True)
                opened = self._perk_menu.open_if_available(perk_ctx)
                if opened and self._replay_recorder is not None:
                    self._replay_recorder.record_perk_menu_open(player_index=0)

        perk_menu_active = self._perk_menu.active

        if not self._paused:
            pulse_delta = dt_ui_ms * (6.0 if self._perk_prompt_hover else -2.0)
            self._perk_prompt_pulse = clamp(self._perk_prompt_pulse + pulse_delta, 0.0, 1000.0)

        prompt_active = perk_pending and (not perk_menu_active) and (not self._paused)
        if prompt_active:
            self._perk_prompt_timer_ms = clamp(self._perk_prompt_timer_ms + dt_ui_ms, 0.0, PERK_PROMPT_MAX_TIMER_MS)
        else:
            self._perk_prompt_timer_ms = clamp(self._perk_prompt_timer_ms - dt_ui_ms, 0.0, PERK_PROMPT_MAX_TIMER_MS)

        self._perk_menu.tick_timeline(dt_ui_ms)

        self._update_lan_wait_gate_debug_override()
        if self._lan_wait_gate_active():
            self._sim_clock.reset()
            return

        dt_world = 0.0 if self._paused or self._perk_menu.active else dt_frame
        if dt_world <= 0.0:
            self._sim_clock.reset()
            # Match legacy transition behavior: keep countdown moving, but at
            # real-time pace while perk-menu transition is holding world ticks.
            self._tick_death_timers(dt_frame, rate=1.0)
            if self._death_transition_ready():
                self._close_failed_run()
            return

        ticks_to_run = self._sim_clock.advance(dt_world)
        if ticks_to_run <= 0:
            return

        dt_tick = float(self._sim_clock.dt_tick)
        input_frame = self._build_local_inputs(dt_frame=dt_frame)
        session = self._sim_session
        if session is None:
            self._tick_death_timers(dt_world)
            if self._death_transition_ready():
                self._close_failed_run()
            return

        session.detail_preset = int(self._deterministic_detail_preset())
        session.fx_toggle = int(self._deterministic_fx_toggle())
        session.spawn_entries = tuple(self._quest.spawn_entries)
        session.spawn_timeline_ms = float(self._quest.spawn_timeline_ms)
        session.no_creatures_timer_ms = float(self._quest.no_creatures_timer_ms)
        session.completion_transition_ms = float(self._quest.completion_transition_ms)

        if self.world.audio_router is not None:
            self.world.audio_router.audio = self.world.audio
            self.world.audio_router.audio_rng = self.world.audio_rng
            self.world.audio_router.demo_mode_active = self.world.demo_mode_active
        if self.world.ground is not None:
            self.world._sync_ground_settings()
            self.world.ground.process_pending()

        for tick_offset in range(int(ticks_to_run)):
            inputs = input_frame if tick_offset == 0 else self._clear_local_input_edges(input_frame)
            recorder = self._replay_recorder
            if recorder is not None:
                tick_index = recorder.record_tick(inputs)
            else:
                tick_index = None
            tick = session.step_tick(
                dt_frame=dt_tick,
                inputs=inputs,
            )
            self.world.apply_step_result(
                tick.step,
                game_tune_started=False,
                apply_audio=True,
                update_camera=True,
            )
            self._quest.spawn_entries = tuple(session.spawn_entries)
            self._quest.spawn_timeline_ms = float(tick.spawn_timeline_ms)
            self._quest.no_creatures_timer_ms = float(tick.no_creatures_timer_ms)
            self._quest.completion_transition_ms = float(tick.completion_transition_ms)
            self._quest.quest_name_timer_ms += float(dt_tick) * 1000.0

            if tick_index is not None:
                world_events = tick.step.events
                self._record_replay_checkpoint(
                    int(tick_index),
                    rng_marks=tick.rng_marks,
                    deaths=world_events.deaths,
                    events=world_events,
                    command_hash=str(tick.step.command_hash),
                )

            if tick.play_hit_sfx:
                self.world.audio_router.play_sfx("sfx_questhit")
            if tick.play_completion_music and self.world.audio is not None:
                play_music(self.world.audio, "crimsonquest")
                playback = self.world.audio.music.playbacks.get("crimsonquest")
                if playback is not None:
                    playback.volume = 0.0
                    try:
                        rl.set_music_volume(playback.music, 0.0)
                    except RuntimeError:
                        playback.volume = 0.0

            if tick.completed:
                if self._outcome is None:
                    fired, hit = shots_from_state(self.state, player_index=int(self.player.index))
                    most_used_weapon_id = most_used_weapon_id_for_player(
                        self.state,
                        player_index=int(self.player.index),
                        fallback_weapon_id=int(self.player.weapon_id),
                    )
                    player_health_values = tuple(float(player.health) for player in self.world.players)
                    player2_health = None
                    if len(player_health_values) >= 2:
                        player2_health = float(player_health_values[1])
                    self._outcome = QuestRunOutcome(
                        kind="completed",
                        level=str(self._quest.level),
                        base_time_ms=int(self._quest.spawn_timeline_ms),
                        player_health=float(player_health_values[0] if player_health_values else self.player.health),
                        player2_health=player2_health,
                        player_health_values=player_health_values,
                        pending_perk_count=int(self.state.perk_selection.pending_count),
                        experience=int(self.player.experience),
                        kill_count=int(self.creatures.kill_count),
                        weapon_id=int(self.player.weapon_id),
                        shots_fired=fired,
                        shots_hit=hit,
                        most_used_weapon_id=int(most_used_weapon_id),
                    )
                self._save_replay()
                self.close_requested = True
                break

            if self._death_transition_ready():
                self._close_failed_run()
                break

    def _update_lan_match(self, *, dt_frame: float, dt_ui_ms: float) -> None:
        runtime = self._lan_runtime
        if runtime is None:
            return
        session = self._sim_session
        if session is None:
            return

        runtime.update()
        role = str(self._lan_role)
        self._consume_net_runtime_recovery(mode_name="quests")
        if str(runtime.error or ""):
            self.close_requested = True
            return
        if self.world.audio_router is not None:
            self.world.audio_router.audio = self.world.audio
            self.world.audio_router.audio_rng = self.world.audio_rng
            self.world.audio_router.demo_mode_active = self.world.demo_mode_active
        if self.world.ground is not None:
            self.world._sync_ground_settings()
            self.world.ground.process_pending()
        self._trace_lan_terrain_generation()
        if bool(self._lan_terrain_generation_pending()):
            self._lan_capture_clock.reset()
            return

        if role == "host" and (not bool(runtime.host_remote_inputs_ready())):
            return

        if bool(self._paused):
            self._sim_clock.reset()
            # Mirror legacy: keep the fail transition timers moving at real-time pace while paused.
            self._tick_death_timers(dt_frame, rate=1.0)
            if self._death_transition_ready():
                self._close_failed_run()
            return

        dt_tick = float(self._lan_capture_clock.dt_tick)

        def _consume_lan_frames() -> bool:
            while True:
                frame = runtime.pop_tick_frame()
                if frame is None:
                    return False

                packed_inputs = list(frame.frame_inputs)
                player_inputs = [unpack_player_input(packed) for packed in packed_inputs]
                recorder = self._replay_recorder
                if recorder is not None:
                    tick_index = recorder.record_tick(player_inputs)
                else:
                    tick_index = None

                session.detail_preset = int(self._deterministic_detail_preset())
                session.fx_toggle = int(self._deterministic_fx_toggle())
                session.spawn_entries = tuple(self._quest.spawn_entries)
                session.spawn_timeline_ms = float(self._quest.spawn_timeline_ms)
                session.no_creatures_timer_ms = float(self._quest.no_creatures_timer_ms)
                session.completion_transition_ms = float(self._quest.completion_transition_ms)

                tick = session.step_tick(
                    dt_frame=float(dt_tick),
                    inputs=player_inputs,
                )

                remote_command_hash = str(frame.command_hash or "")
                remote_state_hash = str(frame.state_hash or "")
                local_command_hash = str(tick.step.command_hash)
                local_state_hash = ""
                if role == "join":
                    if remote_command_hash and remote_command_hash != local_command_hash:
                        runtime.note_desync(
                            kind="command_hash",
                            tick_index=int(frame.tick_index),
                            expected=str(remote_command_hash),
                            actual=str(local_command_hash),
                        )
                    if remote_state_hash:
                        local_state_hash = str(
                            build_checkpoint(
                                tick_index=int(frame.tick_index),
                                world=self.world.world_state,
                                elapsed_ms=float(tick.elapsed_ms),
                                creature_count_override=int(tick.creature_count_world_step),
                            ).state_hash,
                        )
                        if local_state_hash != remote_state_hash:
                            runtime.note_desync(
                                kind="state_hash",
                                tick_index=int(frame.tick_index),
                                expected=str(remote_state_hash),
                                actual=str(local_state_hash),
                            )

                state_hash = ""
                if role == "host":
                    tick_i = int(frame.tick_index)
                    if int(tick_i) < 5 or (int(tick_i) % int(STATE_HASH_PERIOD_TICKS)) == 0:
                        state_hash = str(
                            build_checkpoint(
                                tick_index=int(frame.tick_index),
                                world=self.world.world_state,
                                elapsed_ms=float(tick.elapsed_ms),
                                creature_count_override=int(tick.creature_count_world_step),
                            ).state_hash,
                        )
                self.world.apply_step_result(
                    tick.step,
                    game_tune_started=False,
                    apply_audio=True,
                    update_camera=True,
                )
                self._quest.spawn_entries = tuple(session.spawn_entries)
                self._quest.spawn_timeline_ms = float(tick.spawn_timeline_ms)
                self._quest.no_creatures_timer_ms = float(tick.no_creatures_timer_ms)
                self._quest.completion_transition_ms = float(tick.completion_transition_ms)
                self._quest.quest_name_timer_ms += float(dt_tick) * 1000.0
                self._store_net_runtime_snapshot(
                    mode_name="quests",
                    tick_index=int(frame.tick_index),
                    session_state={
                        "elapsed_ms": float(tick.elapsed_ms),
                        "spawn_timeline_ms": float(tick.spawn_timeline_ms),
                        "no_creatures_timer_ms": float(tick.no_creatures_timer_ms),
                        "completion_transition_ms": float(tick.completion_transition_ms),
                    },
                    mode_state={
                        "quest_spawn_timeline_ms": float(self._quest.spawn_timeline_ms),
                        "quest_no_creatures_timer_ms": float(self._quest.no_creatures_timer_ms),
                        "quest_completion_transition_ms": float(self._quest.completion_transition_ms),
                        "quest_name_timer_ms": float(self._quest.quest_name_timer_ms),
                        "perk_pending_count": int(self.state.perk_selection.pending_count),
                    },
                )

                if tick_index is not None:
                    world_events = tick.step.events
                    self._record_replay_checkpoint(
                        int(tick_index),
                        rng_marks=tick.rng_marks,
                        deaths=world_events.deaths,
                        events=world_events,
                        command_hash=str(tick.step.command_hash),
                    )

                if tick.play_hit_sfx:
                    self.world.audio_router.play_sfx("sfx_questhit")
                if tick.play_completion_music and self.world.audio is not None:
                    play_music(self.world.audio, "crimsonquest")
                    playback = self.world.audio.music.playbacks.get("crimsonquest")
                    if playback is not None:
                        playback.volume = 0.0
                        try:
                            rl.set_music_volume(playback.music, 0.0)
                        except RuntimeError:
                            playback.volume = 0.0

                if role == "host":
                    runtime.broadcast_tick_frame(
                        TickFrame(
                            tick_index=int(frame.tick_index),
                            frame_inputs=list(frame.frame_inputs),
                            command_hash=str(local_command_hash),
                            state_hash=str(state_hash),
                        ),
                    )

                if tick.completed:
                    if self._outcome is None:
                        fired, hit = shots_from_state(self.state, player_index=int(self.player.index))
                        most_used_weapon_id = most_used_weapon_id_for_player(
                            self.state,
                            player_index=int(self.player.index),
                            fallback_weapon_id=int(self.player.weapon_id),
                        )
                        player_health_values = tuple(float(player.health) for player in self.world.players)
                        player2_health = None
                        if len(player_health_values) >= 2:
                            player2_health = float(player_health_values[1])
                        self._outcome = QuestRunOutcome(
                            kind="completed",
                            level=str(self._quest.level),
                            base_time_ms=int(self._quest.spawn_timeline_ms),
                            player_health=float(
                                player_health_values[0] if player_health_values else self.player.health,
                            ),
                            player2_health=player2_health,
                            player_health_values=player_health_values,
                            pending_perk_count=int(self.state.perk_selection.pending_count),
                            experience=int(self.player.experience),
                            kill_count=int(self.creatures.kill_count),
                            weapon_id=int(self.player.weapon_id),
                            shots_fired=fired,
                            shots_hit=hit,
                            most_used_weapon_id=int(most_used_weapon_id),
                        )
                    self._save_replay()
                    self.close_requested = True
                    return True

                if self._death_transition_ready():
                    self._close_failed_run()
                    return True

        if role == "join":
            if _consume_lan_frames():
                return

        ticks_to_capture = self._lan_capture_clock.advance(dt_frame)
        if ticks_to_capture > 0:
            input_frame = self._build_local_inputs(dt_frame=dt_frame)
            # In LAN sessions each peer is a single local player, so always sample
            # inputs using the configured Player 1 bindings (index 0). The network
            # slot mapping is handled by the lockstep runtime.
            local_input_index = 0
            for tick_offset in range(int(ticks_to_capture)):
                inputs = input_frame if tick_offset == 0 else self._clear_local_input_edges(input_frame)
                local_input = PlayerInput()
                if 0 <= local_input_index < len(inputs):
                    local_input = inputs[local_input_index]
                runtime.queue_local_input(pack_player_input(local_input))
        # Pump networking again after queuing local inputs so the host can emit frames
        # in the same render frame (reduces perceived host-side input latency).
        runtime.update()

        _consume_lan_frames()

    def draw(self) -> None:
        perk_menu_active = self._perk_menu.active
        debug_overlay_height = 0.0
        self.world.draw(
            draw_aim_indicators=not perk_menu_active,
            entity_alpha=self._world_entity_alpha(),
        )
        self._draw_screen_fade()

        hud_bottom = 0.0
        if (not perk_menu_active) and self._hud_assets is not None:
            total = int(self._quest.total_spawn_count)
            kills = int(self.creatures.kill_count)
            quest_progress_ratio = float(kills) / float(total) if total > 0 else None
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            hud_bottom = draw_hud_overlay(
                self._hud_assets,
                state=self._hud_state,
                player=self.player,
                players=self.world.players,
                bonus_hud=self.state.bonus_hud,
                elapsed_ms=float(self._quest.spawn_timeline_ms),
                font=self._small,
                frame_dt_ms=self._last_dt_ms,
                show_health=hud_flags.show_health,
                show_weapon=hud_flags.show_weapon,
                show_xp=hud_flags.show_xp,
                show_time=hud_flags.show_time,
                show_quest_hud=hud_flags.show_quest_hud,
                quest_progress_ratio=quest_progress_ratio,
                small_indicators=self._hud_small_indicators(),
                preserve_bugs=bool(self.world.preserve_bugs),
            )

        if debug_enabled() and (not perk_menu_active):
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            god = "on" if self.state.debug_god_mode else "off"
            line = float(self._ui_line_height(scale=0.9))
            self._draw_ui_text(f"debug: [/] weapon  F3 perk+1  F2 god={god}", Vec2(x, y), UI_HINT_COLOR, scale=0.9)
            overlay_end_y = self._draw_lan_debug_info(x=x, y=y + line, line_h=line)
            debug_overlay_height = max(0.0, float(overlay_end_y) - float(y))

        self._draw_quest_title()
        self._draw_quest_complete_banner()

        self._draw_perk_prompt()
        self._perk_menu.draw(self._perk_menu_context())

        if perk_menu_active:
            self._draw_game_cursor()
        elif self._paused:
            self._draw_game_cursor()
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            y += float(debug_overlay_height)
            self._draw_ui_text("paused (TAB)", Vec2(x, y), UI_HINT_COLOR)
        self._draw_lan_wait_overlay()

    def _draw_game_cursor(self) -> None:
        assets = self._perk_menu_assets
        cursor_tex = assets.cursor if assets is not None else None
        mouse_pos = self._ui_mouse
        draw_menu_cursor(
            self.world.particles_texture,
            cursor_tex,
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )

    def _draw_quest_title(self) -> None:
        font = self._grim_mono
        quest = self._quest.quest
        if font is None or quest is None:
            return
        timer_ms = float(self._quest.quest_name_timer_ms)
        if timer_ms <= 0.0 or timer_ms > QUEST_TITLE_TOTAL_MS:
            return
        if timer_ms < QUEST_TITLE_FADE_IN_MS and QUEST_TITLE_FADE_IN_MS > 1e-3:
            alpha = timer_ms / QUEST_TITLE_FADE_IN_MS
        elif timer_ms < (QUEST_TITLE_FADE_IN_MS + QUEST_TITLE_HOLD_MS):
            alpha = 1.0
        else:
            t = timer_ms - (QUEST_TITLE_FADE_IN_MS + QUEST_TITLE_HOLD_MS)
            alpha = max(0.0, 1.0 - (t / max(1e-3, QUEST_TITLE_FADE_OUT_MS)))

        draw_quest_title_overlay(font, quest.title, _quest_level_label(quest.major, quest.minor), alpha=alpha)

    def _draw_quest_complete_banner(self) -> None:
        tex = self._quest_complete_texture
        timer_ms = float(self._quest.completion_transition_ms)
        if tex is None or timer_ms <= 0.0:
            return
        alpha = _quest_complete_banner_alpha(timer_ms)
        if alpha <= 0.0:
            return
        scale = QUEST_COMPLETE_BANNER_SCALE_BASE + timer_ms * QUEST_COMPLETE_BANNER_SCALE_RATE
        width = QUEST_COMPLETE_BANNER_BASE_W * scale
        height = QUEST_COMPLETE_BANNER_BASE_H * scale
        center_x = float(rl.get_screen_width()) * 0.5
        center_y = float(rl.get_screen_height()) * 0.5
        src = rl.Rectangle(0.0, 0.0, float(tex.width), float(tex.height))
        dst = Rect.from_center(Vec2(center_x, center_y), width, height).to_rl()
        tint = rl.Color(255, 255, 255, int(clamp(alpha, 0.0, 1.0) * 255.0))
        rl.draw_texture_pro(tex, src, dst, rl.Vector2(0.0, 0.0), 0.0, tint)



from __future__ import annotations

from ..sim.state_types import PlayerState
from .ids import PerkId


def perk_count_get(player: PlayerState, perk_id: PerkId) -> int:
    idx = int(perk_id)
    if idx < 0:
        return 0
    if idx >= len(player.perk_counts):
        return 0
    return int(player.perk_counts[idx])


def perk_active(player: PlayerState, perk_id: PerkId) -> bool:
    return perk_count_get(player, perk_id) > 0



from __future__ import annotations

from ..ids import PerkId
from ..runtime.apply_context import PerkApplyCtx
from ..runtime.hook_types import PerkHooks


def apply_plaguebearer(ctx: PerkApplyCtx) -> None:
    for player in ctx.players:
        player.plaguebearer_active = True


HOOKS = PerkHooks(
    perk_id=PerkId.PLAGUEBEARER,
    apply_handler=apply_plaguebearer,
)



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import pyray as rl

from grim.geom import Vec2
from grim.math import clamp

from ...effects_atlas import EFFECT_ID_ATLAS_TABLE_BY_ID, SIZE_CODE_GRID, EffectId
from ...perks import PerkId
from ...perks.helpers import perk_active
from ...weapons import WEAPON_BY_ID
from .constants import _RAD_TO_DEG
from .mixin_base import WorldRendererMixinBase

if TYPE_CHECKING:
    from ...sim.state_types import PlayerState

_LAN_PLAYER_RING_RGB: tuple[tuple[int, int, int], ...] = (
    # Match existing trooper torso tint colors for P1/P2.
    (77, 77, 255),
    (255, 140, 89),
    # Distinct colors for P3/P4 (4-player LAN readability).
    (90, 240, 255),
    (255, 120, 230),
)


class WorldRendererTrooperMixin(WorldRendererMixinBase):
    @staticmethod
    def _lan_player_ring_rgb(player_index: int) -> tuple[int, int, int]:
        idx = max(0, min(len(_LAN_PLAYER_RING_RGB) - 1, int(player_index)))
        return _LAN_PLAYER_RING_RGB[idx]

    def _draw_lan_player_ring(
        self,
        *,
        player: PlayerState,
        screen_pos: Vec2,
        base_size: float,
        scale: float,
        alpha: float,
    ) -> None:
        if not bool(self.lan_player_rings_enabled):
            return
        if len(self.players) <= 1:
            return
        if float(player.health) <= 0.0:
            return
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return

        red, green, blue = self._lan_player_ring_rgb(int(player.index))
        outer = max(8.0 * scale, float(base_size) * 0.58)
        thickness = max(2.5 * scale, float(base_size) * 0.11)
        inner = max(0.0, outer - thickness)
        glow_outer = outer + max(2.0 * scale, float(base_size) * 0.08)
        segments = max(24, int(outer * 1.5 + 0.5))
        center = rl.Vector2(screen_pos.x, screen_pos.y)
        core = rl.Color(red, green, blue, int(clamp(alpha * 0.9, 0.0, 1.0) * 255.0 + 0.5))
        glow = rl.Color(red, green, blue, int(clamp(alpha * 0.35, 0.0, 1.0) * 255.0 + 0.5))

        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
        rl.draw_ring(center, inner, outer, 0.0, 360.0, int(segments), core)
        rl.draw_ring(center, outer, glow_outer, 0.0, 360.0, int(segments), glow)
        rl.end_blend_mode()

    def _draw_player_trooper_sprite(
        self,
        texture: rl.Texture,
        player: PlayerState,
        *,
        camera: Vec2,
        view_scale: Vec2,
        scale: float,
        alpha: float = 1.0,
    ) -> None:
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return
        sprite_grid = 8
        cell = float(texture.width) / float(sprite_grid) if sprite_grid > 0 else float(texture.width)
        if cell <= 0.0:
            return

        screen_pos = self._world_to_screen_with(player.pos, camera=camera, view_scale=view_scale)
        base_size = float(player.size) * scale
        base_scale = base_size / cell

        self._draw_lan_player_ring(
            player=player,
            screen_pos=screen_pos,
            base_size=base_size,
            scale=scale,
            alpha=alpha,
        )

        if self.particles_texture is not None and perk_active(player, PerkId.RADIOACTIVE) and alpha > 1e-3:
            atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(EffectId.AURA))
            if atlas is not None:
                aura_grid = SIZE_CODE_GRID.get(int(atlas.size_code))
                if aura_grid:
                    frame = int(atlas.frame)
                    col = frame % aura_grid
                    row = frame // aura_grid
                    cell_w = float(self.particles_texture.width) / float(aura_grid)
                    cell_h = float(self.particles_texture.height) / float(aura_grid)
                    src = rl.Rectangle(
                        cell_w * float(col),
                        cell_h * float(row),
                        max(0.0, cell_w - 2.0),
                        max(0.0, cell_h - 2.0),
                    )
                    t = float(self.elapsed_ms) * 0.001
                    aura_alpha = ((math.sin(t) + 1.0) * 0.1875 + 0.25) * alpha
                    if aura_alpha > 1e-3:
                        size = 100.0 * scale
                        dst = rl.Rectangle(screen_pos.x, screen_pos.y, float(size), float(size))
                        origin = rl.Vector2(size * 0.5, size * 0.5)
                        tint = rl.Color(77, 153, 77, int(clamp(aura_alpha, 0.0, 1.0) * 255.0 + 0.5))
                        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
                        rl.draw_texture_pro(self.particles_texture, src, dst, origin, 0.0, tint)
                        rl.end_blend_mode()

        tint = rl.Color(240, 240, 255, int(255 * alpha + 0.5))
        shadow_tint = rl.Color(0, 0, 0, int(90 * alpha + 0.5))
        overlay_tint = tint
        if len(self.players) > 1:
            index = int(player.index)
            if index == 0:
                overlay_tint = rl.Color(77, 77, 255, tint.a)
            else:
                overlay_tint = rl.Color(255, 140, 89, tint.a)

        def draw(frame: int, *, pos: Vec2, scale_mul: float, rotation: float, color: rl.Color) -> None:
            self._draw_atlas_sprite(
                texture,
                grid=sprite_grid,
                frame=max(0, min(63, int(frame))),
                pos=pos,
                scale=base_scale * float(scale_mul),
                rotation_rad=float(rotation),
                tint=color,
            )

        if player.health > 0.0:
            leg_frame = max(0, min(14, int(player.move_phase + 0.5)))
            torso_frame = leg_frame + 16

            recoil_dir = float(player.aim_heading) + math.pi / 2.0
            recoil = float(player.muzzle_flash_alpha) * 12.0 * scale
            recoil_offset = Vec2.from_polar(recoil_dir, recoil)

            leg_shadow_scale = 1.02
            torso_shadow_scale = 1.03
            leg_shadow_off = 3.0 * scale + base_size * (leg_shadow_scale - 1.0) * 0.5
            torso_shadow_off = 1.0 * scale + base_size * (torso_shadow_scale - 1.0) * 0.5

            draw(
                leg_frame,
                pos=screen_pos.offset(dx=leg_shadow_off, dy=leg_shadow_off),
                scale_mul=leg_shadow_scale,
                rotation=float(player.heading),
                color=shadow_tint,
            )
            draw(
                torso_frame,
                pos=screen_pos.offset(dx=recoil_offset.x + torso_shadow_off, dy=recoil_offset.y + torso_shadow_off),
                scale_mul=torso_shadow_scale,
                rotation=float(player.aim_heading),
                color=shadow_tint,
            )

            draw(
                leg_frame,
                pos=screen_pos,
                scale_mul=1.0,
                rotation=float(player.heading),
                color=tint,
            )
            draw(
                torso_frame,
                pos=screen_pos + recoil_offset,
                scale_mul=1.0,
                rotation=float(player.aim_heading),
                color=overlay_tint,
            )

            if self.particles_texture is not None and float(player.shield_timer) > 1e-3 and alpha > 1e-3:
                atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(EffectId.SHIELD_RING))
                if atlas is not None:
                    shield_grid = SIZE_CODE_GRID.get(int(atlas.size_code))
                    if shield_grid:
                        frame = int(atlas.frame)
                        col = frame % shield_grid
                        row = frame // shield_grid
                        cell_w = float(self.particles_texture.width) / float(shield_grid)
                        cell_h = float(self.particles_texture.height) / float(shield_grid)
                        src = rl.Rectangle(
                            cell_w * float(col),
                            cell_h * float(row),
                            max(0.0, cell_w - 2.0),
                            max(0.0, cell_h - 2.0),
                        )
                        t = float(self.elapsed_ms) * 0.001
                        timer = float(player.shield_timer)
                        strength = (math.sin(t) + 1.0) * 0.25 + timer
                        if timer < 1.0:
                            strength *= timer
                            strength = min(1.0, strength) * alpha
                            if strength > 1e-3:
                                offset_dir = float(player.aim_heading) - math.pi / 2.0
                                center = screen_pos + Vec2.from_polar(offset_dir, 3.0 * scale)

                                half = math.sin(t * 3.0) + 17.5
                                size = half * 2.0 * scale
                                a = int(clamp(strength * 0.4, 0.0, 1.0) * 255.0 + 0.5)
                                tint = rl.Color(91, 180, 255, a)
                                dst = rl.Rectangle(center.x, center.y, float(size), float(size))
                                origin = rl.Vector2(size * 0.5, size * 0.5)
                                rotation_deg = float((t + t) * _RAD_TO_DEG)

                                half = math.sin(t * 3.0) * 4.0 + 24.0
                                size2 = half * 2.0 * scale
                                a2 = int(clamp(strength * 0.3, 0.0, 1.0) * 255.0 + 0.5)
                                tint2 = rl.Color(91, 180, 255, a2)
                                dst2 = rl.Rectangle(center.x, center.y, float(size2), float(size2))
                                origin2 = rl.Vector2(size2 * 0.5, size2 * 0.5)
                                rotation2_deg = float((t * -2.0) * _RAD_TO_DEG)

                                rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
                                rl.draw_texture_pro(self.particles_texture, src, dst, origin, rotation_deg, tint)
                                rl.draw_texture_pro(self.particles_texture, src, dst2, origin2, rotation2_deg, tint2)
                                rl.end_blend_mode()

            if self.muzzle_flash_texture is not None and float(player.muzzle_flash_alpha) > 1e-3 and alpha > 1e-3:
                weapon = WEAPON_BY_ID.get(int(player.weapon_id))
                flags = int(weapon.flags) if weapon is not None and weapon.flags is not None else 0
                if (flags & 0x8) == 0:
                    flash_alpha = clamp(float(player.muzzle_flash_alpha) * 0.8, 0.0, 1.0) * alpha
                    if flash_alpha > 1e-3:
                        size = base_size * (0.5 if (flags & 0x4) else 1.0)
                        heading = float(player.aim_heading) + math.pi / 2.0
                        offset = (float(player.muzzle_flash_alpha) * 12.0 - 21.0) * scale
                        flash_pos = screen_pos + Vec2.from_angle(heading) * offset
                        src = rl.Rectangle(
                            0.0,
                            0.0,
                            float(self.muzzle_flash_texture.width),
                            float(self.muzzle_flash_texture.height),
                        )
                        dst = rl.Rectangle(flash_pos.x, flash_pos.y, size, size)
                        origin = rl.Vector2(size * 0.5, size * 0.5)
                        tint_flash = rl.Color(255, 255, 255, int(flash_alpha * 255.0 + 0.5))
                        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
                        rl.draw_texture_pro(
                            self.muzzle_flash_texture,
                            src,
                            dst,
                            origin,
                            float(player.aim_heading * _RAD_TO_DEG),
                            tint_flash,
                        )
                        rl.end_blend_mode()
            return

        if player.death_timer >= 0.0:
            # Matches the observed frame ramp (32..52) in player_sprite_trace.jsonl.
            frame = 32 + int((16.0 - float(player.death_timer)) * 1.25)
            if frame > 52:
                frame = 52
            if frame < 32:
                frame = 32
        else:
            frame = 52

        dead_shadow_scale = 1.03
        dead_shadow_off = 1.0 * scale + base_size * (dead_shadow_scale - 1.0) * 0.5
        draw(
            frame,
            pos=screen_pos.offset(dx=dead_shadow_off, dy=dead_shadow_off),
            scale_mul=dead_shadow_scale,
            rotation=float(player.aim_heading),
            color=shadow_tint,
        )
        draw(frame, pos=screen_pos, scale_mul=1.0, rotation=float(player.aim_heading), color=overlay_tint)



from __future__ import annotations

from dataclasses import dataclass, field

from grim.geom import Vec2


@dataclass(frozen=True, slots=True)
class PlayerInput:
    move: Vec2 = field(default_factory=Vec2)
    aim: Vec2 = field(default_factory=Vec2)
    fire_down: bool = False
    fire_pressed: bool = False
    reload_pressed: bool = False
    move_to_cursor_pressed: bool = False
    move_forward_pressed: bool | None = None
    move_backward_pressed: bool | None = None
    turn_left_pressed: bool | None = None
    turn_right_pressed: bool | None = None



from __future__ import annotations

from .adapter import ClientLanAdapter, HostLanAdapter
from .lobby import ClientLobby, HostLobby
from .lockstep import ClientLockstepState, HostLockstepState
from .net_runtime import NetRuntime, NetRuntimeConfig
from .protocol import (
    DEFAULT_PORT,
    INPUT_DELAY_TICKS,
    MAX_PLAYERS,
    PROTOCOL_VERSION,
    RELIABLE_RESEND_MS,
    STATE_HASH_PERIOD_TICKS,
    TICK_RATE,
)
from .relay_service import RelayServer, RelayServerConfig
from .reliable import ReliableLink
from .resync import ResyncAssembler, ResyncBuildError, build_resync_messages
from .transport import PeerAddr, UdpTransport

__all__ = [
    "ClientLanAdapter",
    "ClientLobby",
    "ClientLockstepState",
    "DEFAULT_PORT",
    "HostLanAdapter",
    "HostLobby",
    "HostLockstepState",
    "INPUT_DELAY_TICKS",
    "MAX_PLAYERS",
    "NetRuntime",
    "NetRuntimeConfig",
    "PROTOCOL_VERSION",
    "PeerAddr",
    "RELIABLE_RESEND_MS",
    "ReliableLink",
    "RelayServer",
    "RelayServerConfig",
    "ResyncAssembler",
    "ResyncBuildError",
    "STATE_HASH_PERIOD_TICKS",
    "TICK_RATE",
    "UdpTransport",
    "build_resync_messages",
]



from __future__ import annotations

from ..helpers import perk_count_get
from ..ids import PerkId
from ..runtime.effects_context import PerksUpdateEffectsCtx
from ..runtime.hook_types import PerkHooks


def update_lean_mean_exp_machine(ctx: PerksUpdateEffectsCtx) -> None:
    ctx.state.lean_mean_exp_timer -= ctx.dt
    if ctx.state.lean_mean_exp_timer < 0.0:
        ctx.state.lean_mean_exp_timer = 0.25
        if not ctx.players:
            return

        # Native `perks_update_effects` uses global `perk_count_get` and awards the
        # periodic XP tick only to player 0 (`player_experience[0]`).
        player0 = ctx.players[0]
        perk_count = perk_count_get(player0, PerkId.LEAN_MEAN_EXP_MACHINE)
        if perk_count > 0:
            player0.experience += perk_count * 10


HOOKS = PerkHooks(
    perk_id=PerkId.LEAN_MEAN_EXP_MACHINE,
    effects_steps=(update_lean_mean_exp_machine,),
)



from __future__ import annotations

from collections.abc import Sequence

from ..math_parity import f32
from ..perks.helpers import perk_active
from ..projectiles import CreatureDamageApplier, Damageable
from ..sim.state_types import BonusPickupEvent, GameplayState, PlayerState
from .apply import bonus_apply
from .hud import bonus_hud_update
from .ids import BonusId
from .pool import BONUS_PICKUP_LINGER, BONUS_TELEKINETIC_PICKUP_MS, bonus_find_aim_hover_entry

_REFLEX_TIMER_SUBTRACT_BIAS = 4e-9


def bonus_telekinetic_update(
    state: GameplayState,
    players: list[PlayerState],
    dt: float,
    *,
    creatures: Sequence[Damageable] | None = None,
    apply_creature_damage: CreatureDamageApplier | None = None,
    detail_preset: int = 5,
    defer_freeze_corpse_fx: bool = False,
    freeze_corpse_indices: set[int] | None = None,
) -> list[BonusPickupEvent]:
    """Allow Telekinetic perk owners to pick up bonuses by aiming at them."""
    from ..perks import PerkId

    if dt <= 0.0:
        return []

    pickups: list[BonusPickupEvent] = []
    dt_ms = float(dt) * 1000.0

    for player in players:
        if player.health <= 0.0:
            continue

        hovered = bonus_find_aim_hover_entry(player, state.bonus_pool)
        if hovered is None:
            player.bonus_aim_hover_index = -1
            player.bonus_aim_hover_timer_ms = 0.0
            continue

        idx, entry = hovered
        player.bonus_aim_hover_index = int(idx)
        player.bonus_aim_hover_timer_ms += dt_ms

        if player.bonus_aim_hover_timer_ms <= BONUS_TELEKINETIC_PICKUP_MS:
            continue
        if not perk_active(player, PerkId.TELEKINETIC):
            continue
        if entry.picked or entry.bonus_id == 0:
            continue

        bonus_apply(
            state,
            player,
            BonusId(int(entry.bonus_id)),
            amount=int(entry.amount),
            origin=entry,
            creatures=creatures,
            players=players,
            apply_creature_damage=apply_creature_damage,
            detail_preset=int(detail_preset),
            defer_freeze_corpse_fx=bool(defer_freeze_corpse_fx),
            freeze_corpse_indices=freeze_corpse_indices,
        )
        entry.picked = True
        entry.time_left = BONUS_PICKUP_LINGER
        pickups.append(
            BonusPickupEvent(
                player_index=int(player.index),
                bonus_id=int(entry.bonus_id),
                amount=int(entry.amount),
                pos=entry.pos,
            ),
        )

        # Match the exe: after a telekinetic pickup, reset the hover accumulator.
        player.bonus_aim_hover_index = -1
        player.bonus_aim_hover_timer_ms = 0.0
        break

    return pickups


def bonus_update(
    state: GameplayState,
    players: list[PlayerState],
    dt: float,
    *,
    creatures: Sequence[Damageable] | None = None,
    update_hud: bool = True,
    apply_creature_damage: CreatureDamageApplier | None = None,
    detail_preset: int = 5,
    defer_freeze_corpse_fx: bool = False,
    freeze_corpse_indices: set[int] | None = None,
) -> list[BonusPickupEvent]:
    """Advance world bonuses and global timers (subset of `bonus_update`)."""

    pickups = bonus_telekinetic_update(
        state,
        players,
        dt,
        creatures=creatures,
        apply_creature_damage=apply_creature_damage,
        detail_preset=int(detail_preset),
        defer_freeze_corpse_fx=bool(defer_freeze_corpse_fx),
        freeze_corpse_indices=freeze_corpse_indices,
    )
    pickups.extend(
        state.bonus_pool.update(
            dt,
            state=state,
            players=players,
            creatures=creatures,
            apply_creature_damage=apply_creature_damage,
            detail_preset=int(detail_preset),
            defer_freeze_corpse_fx=bool(defer_freeze_corpse_fx),
            freeze_corpse_indices=freeze_corpse_indices,
        ),
    )

    if dt > 0.0:
        # Native `bonus_update` decrements Freeze + Double XP here; other global
        # timers are advanced earlier in the gameplay loop.
        double_xp = float(state.bonuses.double_experience)
        if double_xp <= 0.0:
            state.bonuses.double_experience = 0.0
        else:
            state.bonuses.double_experience = float(f32(float(double_xp) - float(dt)))

        freeze = float(state.bonuses.freeze)
        if freeze <= 0.0:
            state.bonuses.freeze = 0.0
        else:
            state.bonuses.freeze = float(f32(float(freeze) - float(dt)))

    if update_hud:
        bonus_hud_update(state, players, dt=dt)

    return pickups


def bonus_update_pre_pickup_timers(state: GameplayState, dt: float) -> None:
    """Advance global timers that native decrements before `bonus_update`."""

    if dt <= 0.0:
        return
    if float(state.bonuses.weapon_power_up) > 0.0:
        state.bonuses.weapon_power_up = float(f32(float(state.bonuses.weapon_power_up) - float(dt)))
    if float(state.bonuses.energizer) > 0.0:
        state.bonuses.energizer = float(f32(float(state.bonuses.energizer) - float(dt)))
    if float(state.bonuses.reflex_boost) > 0.0:
        reflex_before = float(state.bonuses.reflex_boost)
        subtract = float(dt)
        if 0.0 < reflex_before < 1.0:
            # Native x87 timer math trends slightly lower than straight f32 subtraction in this window.
            subtract += float(_REFLEX_TIMER_SUBTRACT_BIAS)
        state.bonuses.reflex_boost = float(f32(float(reflex_before) - float(subtract)))



from __future__ import annotations

from ..perks import PerkId
from ..perks.helpers import perk_active
from ..weapon_runtime.assign import weapon_assign_player
from .apply_context import BonusApplyCtx


def apply_weapon(ctx: BonusApplyCtx) -> None:
    weapon_id = int(ctx.amount)
    if perk_active(ctx.player, PerkId.ALTERNATE_WEAPON) and ctx.player.alt_weapon_id is None:
        ctx.player.alt_weapon_id = int(ctx.player.weapon_id)
        ctx.player.alt_clip_size = int(ctx.player.clip_size)
        ctx.player.alt_ammo = float(ctx.player.ammo)
        ctx.player.alt_reload_active = bool(ctx.player.reload_active)
        ctx.player.alt_reload_timer = float(ctx.player.reload_timer)
        ctx.player.alt_shot_cooldown = float(ctx.player.shot_cooldown)
        ctx.player.alt_reload_timer_max = float(ctx.player.reload_timer_max)
    weapon_assign_player(ctx.player, weapon_id, state=ctx.state)



from __future__ import annotations

from dataclasses import dataclass, field

from .relay_protocol import RELIABLE_RESEND_MS, NetMessage, Packet


@dataclass(slots=True)
class _PendingReliable:
    packet: Packet
    sent_at_ms: int


@dataclass(slots=True)
class RelayReliableLink:
    """Per-peer reliability state for relay protocol packets."""

    resend_ms: int = RELIABLE_RESEND_MS
    _next_seq: int = 1
    _recv_highest_seq: int = 0
    _pending: dict[int, _PendingReliable] = field(default_factory=dict)
    _recv_buffer: dict[int, Packet] = field(default_factory=dict)
    _rtt_last_ms: int = 0
    _rtt_ewma_ms: float = 0.0
    _resend_count: int = 0

    @property
    def recv_highest_seq(self) -> int:
        return int(self._recv_highest_seq)

    @property
    def pending_count(self) -> int:
        return int(len(self._pending))

    @property
    def rtt_last_ms(self) -> int:
        return int(self._rtt_last_ms)

    @property
    def rtt_ewma_ms(self) -> float:
        return float(self._rtt_ewma_ms)

    @property
    def resend_count(self) -> int:
        return int(self._resend_count)

    def prime_recv_seq(self, seq: int) -> None:
        value = int(seq)
        if value <= int(self._recv_highest_seq):
            return
        self._recv_highest_seq = int(value)

    def build_packet(self, message: NetMessage, *, reliable: bool, now_ms: int) -> Packet:
        seq = 0
        if reliable:
            seq = int(self._next_seq)
            self._next_seq += 1
        packet = Packet(
            seq=int(seq),
            ack=int(self._recv_highest_seq),
            reliable=bool(reliable),
            message=message,
        )
        if reliable:
            self._pending[int(seq)] = _PendingReliable(packet=packet, sent_at_ms=int(now_ms))
        return packet

    def ingest_packet(self, packet: Packet, *, now_ms: int) -> tuple[list[NetMessage], bool]:
        self._apply_ack(int(packet.ack), now_ms=int(now_ms))
        if not bool(packet.reliable):
            return [packet.message], False

        seq = int(packet.seq)
        if seq <= 0:
            return [], False

        if seq <= int(self._recv_highest_seq):
            return [], True
        if seq in self._recv_buffer:
            return [], True

        self._recv_buffer[int(seq)] = packet

        delivered: list[NetMessage] = []
        next_seq = int(self._recv_highest_seq) + 1
        while next_seq in self._recv_buffer:
            next_packet = self._recv_buffer.pop(int(next_seq))
            delivered.append(next_packet.message)
            self._recv_highest_seq = int(next_seq)
            next_seq += 1

        return delivered, False

    def _apply_ack(self, ack: int, *, now_ms: int) -> None:
        if int(ack) <= 0:
            return
        to_drop = [seq for seq in self._pending if int(seq) <= int(ack)]
        if not to_drop:
            return

        newest_seq = max(int(seq) for seq in to_drop)
        pending = self._pending.get(int(newest_seq))
        if pending is not None:
            sample_ms = max(0, int(now_ms) - int(pending.sent_at_ms))
            self._rtt_last_ms = int(sample_ms)
            if self._rtt_ewma_ms <= 0.0:
                self._rtt_ewma_ms = float(sample_ms)
            else:
                self._rtt_ewma_ms = float(self._rtt_ewma_ms * 0.9 + float(sample_ms) * 0.1)

        for seq in to_drop:
            self._pending.pop(int(seq), None)

    def poll_resends(self, *, now_ms: int) -> list[Packet]:
        out: list[Packet] = []
        for seq, pending in list(self._pending.items()):
            if int(now_ms) - int(pending.sent_at_ms) < int(self.resend_ms):
                continue
            refreshed = Packet(
                seq=int(pending.packet.seq),
                ack=int(self._recv_highest_seq),
                reliable=True,
                message=pending.packet.message,
            )
            self._pending[int(seq)] = _PendingReliable(packet=refreshed, sent_at_ms=int(now_ms))
            out.append(refreshed)
            self._resend_count += 1
        return out



from __future__ import annotations

import math

import pyray as rl

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp

from ...effects import ParticleStyleId
from ...effects_atlas import EFFECT_ID_ATLAS_TABLE_BY_ID, SIZE_CODE_GRID, EffectId
from .constants import _RAD_TO_DEG
from .mixin_base import WorldRendererMixinBase


class WorldRendererEffectsMixin(WorldRendererMixinBase):
    def _draw_particle_pool(self, *, camera: Vec2, view_scale: Vec2, alpha: float = 1.0) -> None:
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return
        texture = self.particles_texture
        if texture is None:
            return

        particles = self.state.particles.entries
        if not any(entry.active for entry in particles):
            return

        scale = self._view_scale_avg(view_scale)

        def src_rect(effect_id: int) -> rl.Rectangle | None:
            atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(effect_id))
            if atlas is None:
                return None
            grid = SIZE_CODE_GRID.get(int(atlas.size_code))
            if not grid:
                return None
            frame = int(atlas.frame)
            col = frame % grid
            row = frame // grid
            cell_w = float(texture.width) / float(grid)
            cell_h = float(texture.height) / float(grid)
            return rl.Rectangle(
                cell_w * float(col),
                cell_h * float(row),
                max(0.0, cell_w - 2.0),
                max(0.0, cell_h - 2.0),
            )

        src_large = src_rect(13)
        src_normal = src_rect(12)
        src_style_8 = src_rect(2)
        if src_normal is None or src_style_8 is None:
            return

        fx_detail_1 = self.config.fx_detail(level=1, default=True) if self.config is not None else True

        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)

        if fx_detail_1 and src_large is not None:
            alpha_byte = int(clamp(alpha * 0.065, 0.0, 1.0) * 255.0 + 0.5)
            tint = rl.Color(255, 255, 255, alpha_byte)
            for idx, entry in enumerate(particles):
                if not entry.active or (idx % 2) or int(entry.style_id) == int(ParticleStyleId.BUBBLEGUN):
                    continue
                radius = (math.sin((1.0 - float(entry.intensity)) * 1.5707964) + 0.1) * 55.0 + 4.0
                radius = max(radius, 16.0)
                size = max(0.0, radius * 2.0 * scale)
                if size <= 0.0:
                    continue
                screen = self._world_to_screen_with(entry.pos, camera=camera, view_scale=view_scale)
                dst = rl.Rectangle(screen.x, screen.y, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                rl.draw_texture_pro(texture, src_large, dst, origin, 0.0, tint)

        for entry in particles:
            if not entry.active or int(entry.style_id) == int(ParticleStyleId.BUBBLEGUN):
                continue
            radius = math.sin((1.0 - float(entry.intensity)) * 1.5707964) * 24.0
            if int(entry.style_id) == int(ParticleStyleId.BLOW_TORCH):
                radius *= 0.8
            radius = max(radius, 2.0)
            size = max(0.0, radius * 2.0 * scale)
            if size <= 0.0:
                continue
            screen = self._world_to_screen_with(entry.pos, camera=camera, view_scale=view_scale)
            dst = rl.Rectangle(screen.x, screen.y, size, size)
            origin = rl.Vector2(size * 0.5, size * 0.5)
            rotation_deg = float(entry.spin) * _RAD_TO_DEG
            tint = RGBA(entry.scale_x, entry.scale_y, entry.scale_z, float(entry.age) * alpha).to_rl()
            rl.draw_texture_pro(texture, src_normal, dst, origin, rotation_deg, tint)

        alpha_byte = int(clamp(alpha, 0.0, 1.0) * 255.0 + 0.5)
        for entry in particles:
            if not entry.active or int(entry.style_id) != int(ParticleStyleId.BUBBLEGUN):
                continue
            wobble = math.sin(float(entry.spin)) * 3.0
            half_h = (wobble + 15.0) * float(entry.scale_x) * 7.0
            half_w = (15.0 - wobble) * float(entry.scale_x) * 7.0
            w = max(0.0, half_w * 2.0 * scale)
            h = max(0.0, half_h * 2.0 * scale)
            if w <= 0.0 or h <= 0.0:
                continue
            screen = self._world_to_screen_with(entry.pos, camera=camera, view_scale=view_scale)
            dst = rl.Rectangle(screen.x, screen.y, w, h)
            origin = rl.Vector2(w * 0.5, h * 0.5)
            tint = rl.Color(255, 255, 255, int(float(entry.age) * alpha_byte + 0.5))
            rl.draw_texture_pro(texture, src_style_8, dst, origin, 0.0, tint)

        rl.end_blend_mode()

    def _draw_sprite_effect_pool(
        self,
        *,
        camera: Vec2,
        view_scale: Vec2,
        alpha: float = 1.0,
    ) -> None:
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return
        if self.config is not None and not self.config.fx_detail(level=2, default=False):
            return
        texture = self.particles_texture
        if texture is None:
            return

        effects = self.state.sprite_effects.entries
        if not any(entry.active for entry in effects):
            return

        atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(EffectId.EXPLOSION_PUFF))
        if atlas is None:
            return
        grid = SIZE_CODE_GRID.get(int(atlas.size_code))
        if not grid:
            return
        frame = int(atlas.frame)
        col = frame % grid
        row = frame // grid
        cell_w = float(texture.width) / float(grid)
        cell_h = float(texture.height) / float(grid)
        src = rl.Rectangle(cell_w * float(col), cell_h * float(row), cell_w, cell_h)
        scale = self._view_scale_avg(view_scale)

        rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)
        for entry in effects:
            if not entry.active:
                continue
            size = float(entry.scale) * scale
            if size <= 0.0:
                continue
            screen = self._world_to_screen_with(entry.pos, camera=camera, view_scale=view_scale)
            dst = rl.Rectangle(screen.x, screen.y, size, size)
            origin = rl.Vector2(size * 0.5, size * 0.5)
            rotation_deg = float(entry.rotation) * _RAD_TO_DEG
            tint = entry.color.scaled_alpha(alpha).to_rl()
            rl.draw_texture_pro(texture, src, dst, origin, rotation_deg, tint)
        rl.end_blend_mode()

    def _draw_effect_pool(self, *, camera: Vec2, view_scale: Vec2, alpha: float = 1.0) -> None:
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return
        texture = self.particles_texture
        if texture is None:
            return

        effects = self.state.effects.entries
        if not any(entry.flags and entry.age >= 0.0 for entry in effects):
            return

        scale = self._view_scale_avg(view_scale)

        src_cache: dict[int, rl.Rectangle] = {}

        def src_rect(effect_id: int) -> rl.Rectangle | None:
            cached = src_cache.get(effect_id)
            if cached is not None:
                return cached

            atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(effect_id))
            if atlas is None:
                return None
            grid = SIZE_CODE_GRID.get(int(atlas.size_code))
            if not grid:
                return None
            frame = int(atlas.frame)
            col = frame % grid
            row = frame // grid
            cell_w = float(texture.width) / float(grid)
            cell_h = float(texture.height) / float(grid)
            # Native effect pool clamps UVs to (cell_size - 2px) to avoid bleeding.
            src = rl.Rectangle(
                cell_w * float(col),
                cell_h * float(row),
                max(0.0, cell_w - 2.0),
                max(0.0, cell_h - 2.0),
            )
            src_cache[effect_id] = src
            return src

        def draw_entry(entry: object) -> None:
            effect_id = int(getattr(entry, "effect_id", 0))
            src = src_rect(effect_id)
            if src is None:
                return

            pos = getattr(entry, "pos", None)
            if not isinstance(pos, Vec2):
                return
            screen = self._world_to_screen_with(pos, camera=camera, view_scale=view_scale)

            half_w = float(getattr(entry, "half_width", 0.0))
            half_h = float(getattr(entry, "half_height", 0.0))
            local_scale = float(getattr(entry, "scale", 1.0))
            w = max(0.0, half_w * 2.0 * local_scale * scale)
            h = max(0.0, half_h * 2.0 * local_scale * scale)
            if w <= 0.0 or h <= 0.0:
                return

            rotation_deg = float(getattr(entry, "rotation", 0.0)) * _RAD_TO_DEG
            raw_color = getattr(entry, "color", None)
            if not isinstance(raw_color, RGBA):
                return
            tint = raw_color.scaled_alpha(alpha).to_rl()

            dst = rl.Rectangle(screen.x, screen.y, float(w), float(h))
            origin = rl.Vector2(float(w) * 0.5, float(h) * 0.5)
            rl.draw_texture_pro(texture, src, dst, origin, rotation_deg, tint)

        rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)
        for entry in effects:
            if not entry.flags or entry.age < 0.0:
                continue
            if int(entry.flags) & 0x40:
                draw_entry(entry)
        rl.end_blend_mode()

        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
        for entry in effects:
            if not entry.flags or entry.age < 0.0:
                continue
            if not (int(entry.flags) & 0x40):
                draw_entry(entry)
        rl.end_blend_mode()



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field

from ..replay.types import PackedPlayerInput
from .relay_protocol import RbInputBatch, RbInputSample

_NEUTRAL_INPUT: PackedPlayerInput = [0.0, 0.0, [0.0, 0.0], 0]
_MAX_SENT_HISTORY_TICKS = 256
_MAX_RESEND_SAMPLES = 64


@dataclass(frozen=True, slots=True)
class RollbackFrame:
    tick_index: int
    frame_inputs: list[PackedPlayerInput]
    predicted_slots: tuple[int, ...] = ()


@dataclass(slots=True)
class RollbackController:
    """Input timeline + prediction + rollback trigger bookkeeping.

    This controller does not mutate game state directly. It emits canonical tick
    inputs and reports rollback windows when late remote inputs disagree with
    previously-emitted predicted inputs.
    """

    player_count: int
    local_slot_index: int
    input_delay_ticks: int = 1
    max_rollback_ticks: int = 8
    max_sent_history_ticks: int = _MAX_SENT_HISTORY_TICKS
    max_resend_samples: int = _MAX_RESEND_SAMPLES

    _capture_tick: int = 0
    _next_emit_tick: int = 0
    _known_by_slot: dict[int, dict[int, PackedPlayerInput]] = field(default_factory=dict)
    _emitted_frames: dict[int, list[PackedPlayerInput]] = field(default_factory=dict)
    _pending_frames: deque[RollbackFrame] = field(default_factory=deque)
    _pending_rollback_from: int | None = None
    _pending_resync_from: int | None = None

    rollback_count: int = 0
    prediction_mismatches: int = 0
    max_rollback_distance: int = 0

    def __post_init__(self) -> None:
        players = max(1, int(self.player_count))
        self.player_count = int(players)
        slot = max(0, min(int(self.local_slot_index), int(players) - 1))
        self.local_slot_index = int(slot)
        self.input_delay_ticks = max(0, int(self.input_delay_ticks))
        self.max_rollback_ticks = max(1, int(self.max_rollback_ticks))
        for slot_index in range(int(self.player_count)):
            self._known_by_slot[int(slot_index)] = {}

    @property
    def capture_tick(self) -> int:
        return int(self._capture_tick)

    @property
    def next_emit_tick(self) -> int:
        return int(self._next_emit_tick)

    def queue_local_input(self, packed_input: PackedPlayerInput) -> RbInputBatch:
        target_tick = int(self._capture_tick) + int(self.input_delay_ticks)
        self._known_by_slot[int(self.local_slot_index)][int(target_tick)] = list(packed_input)

        history = max(1, int(self.max_sent_history_ticks))
        oldest = max(int(self._next_emit_tick), int(target_tick) - int(history) + 1)
        max_samples = max(1, int(self.max_resend_samples))

        samples: list[RbInputSample] = []
        slot_map = self._known_by_slot[int(self.local_slot_index)]
        count = 0
        for tick in range(int(target_tick), int(oldest) - 1, -1):
            value = slot_map.get(int(tick))
            if value is None:
                continue
            samples.append(RbInputSample(tick_index=int(tick), packed_input=list(value)))
            count += 1
            if count >= int(max_samples):
                break

        for tick in list(slot_map.keys()):
            if int(tick) < int(oldest) or int(tick) > int(target_tick):
                slot_map.pop(int(tick), None)

        self._capture_tick += 1
        self._emit_ready_frames()
        return RbInputBatch(slot_index=int(self.local_slot_index), samples=samples)

    def ingest_remote_samples(self, *, slot_index: int, samples: list[RbInputSample]) -> None:
        slot = int(slot_index)
        if slot < 0 or slot >= int(self.player_count):
            return
        if slot == int(self.local_slot_index):
            return
        known = self._known_by_slot[int(slot)]
        for sample in samples:
            tick = int(sample.tick_index)
            if tick < 0:
                continue
            incoming = list(sample.packed_input)
            prev = known.get(int(tick))
            known[int(tick)] = incoming
            if prev is not None and prev == incoming:
                continue
            self._note_late_mismatch(slot_index=int(slot), tick_index=int(tick), incoming=incoming)
        self._emit_ready_frames()

    def pop_frame(self) -> RollbackFrame | None:
        if not self._pending_frames:
            return None
        return self._pending_frames.popleft()

    def drain_rollback_from(self) -> int | None:
        value = self._pending_rollback_from
        self._pending_rollback_from = None
        return None if value is None else int(value)

    def drain_resync_from(self) -> int | None:
        value = self._pending_resync_from
        self._pending_resync_from = None
        return None if value is None else int(value)

    def rebuild_emitted_from(self, from_tick: int) -> list[RollbackFrame]:
        start = max(0, int(from_tick))
        head = int(self._next_emit_tick)
        if start >= int(head):
            return []

        rebuilt: list[RollbackFrame] = []
        for tick in range(int(start), int(head)):
            predicted_slots: list[int] = []
            frame_inputs: list[PackedPlayerInput] = []
            for slot in range(int(self.player_count)):
                value = self._known_for_tick(int(slot), int(tick))
                if value is None:
                    predicted_slots.append(int(slot))
                    value = self._predict_for_tick(int(slot), int(tick))
                frame_inputs.append(list(value))
            self._emitted_frames[int(tick)] = [list(item) for item in frame_inputs]
            rebuilt.append(
                RollbackFrame(
                    tick_index=int(tick),
                    frame_inputs=[list(item) for item in frame_inputs],
                    predicted_slots=tuple(predicted_slots),
                ),
            )
        return rebuilt

    def reset_to_tick(self, tick_index: int) -> None:
        target = max(0, int(tick_index))
        self._capture_tick = max(int(self._capture_tick), int(target))
        self._next_emit_tick = max(int(self._next_emit_tick), int(target))
        self._pending_frames.clear()
        self._pending_rollback_from = None
        self._pending_resync_from = None
        for slot_map in self._known_by_slot.values():
            for tick in list(slot_map.keys()):
                if int(tick) >= int(target):
                    continue
                slot_map.pop(int(tick), None)
        for tick in list(self._emitted_frames.keys()):
            if int(tick) >= int(target):
                continue
            self._emitted_frames.pop(int(tick), None)

    def _emit_ready_frames(self) -> None:
        local_known = self._known_by_slot[int(self.local_slot_index)]
        while int(self._next_emit_tick) in local_known:
            tick = int(self._next_emit_tick)
            predicted_slots: list[int] = []
            frame_inputs: list[PackedPlayerInput] = []
            for slot in range(int(self.player_count)):
                value = self._known_for_tick(int(slot), int(tick))
                if value is None:
                    predicted_slots.append(int(slot))
                    value = self._predict_for_tick(int(slot), int(tick))
                frame_inputs.append(list(value))
            self._pending_frames.append(
                RollbackFrame(
                    tick_index=int(tick),
                    frame_inputs=[list(item) for item in frame_inputs],
                    predicted_slots=tuple(predicted_slots),
                ),
            )
            self._emitted_frames[int(tick)] = [list(item) for item in frame_inputs]
            self._next_emit_tick += 1
            self._prune_emitted()

    def _known_for_tick(self, slot: int, tick: int) -> PackedPlayerInput | None:
        return self._known_by_slot[int(slot)].get(int(tick))

    def _predict_for_tick(self, slot: int, tick: int) -> PackedPlayerInput:
        known = self._known_by_slot[int(slot)]
        for back in range(int(tick) - 1, -1, -1):
            value = known.get(int(back))
            if value is not None:
                return list(value)
        return list(_NEUTRAL_INPUT)

    def _note_late_mismatch(self, *, slot_index: int, tick_index: int, incoming: PackedPlayerInput) -> None:
        emitted = self._emitted_frames.get(int(tick_index))
        if emitted is None:
            return
        if int(slot_index) >= len(emitted):
            return
        predicted = emitted[int(slot_index)]
        if list(predicted) == list(incoming):
            return

        self.prediction_mismatches = int(self.prediction_mismatches) + 1
        distance = max(0, int(self._next_emit_tick) - int(tick_index))
        self.max_rollback_distance = max(int(self.max_rollback_distance), int(distance))

        if int(distance) > int(self.max_rollback_ticks):
            pending = self._pending_resync_from
            if pending is None:
                self._pending_resync_from = int(tick_index)
            else:
                self._pending_resync_from = min(int(pending), int(tick_index))
            return

        self.rollback_count = int(self.rollback_count) + 1
        pending = self._pending_rollback_from
        if pending is None:
            self._pending_rollback_from = int(tick_index)
        else:
            self._pending_rollback_from = min(int(pending), int(tick_index))

    def _prune_emitted(self) -> None:
        keep_from = int(self._next_emit_tick) - max(1, int(self.max_rollback_ticks)) - 2
        for tick in list(self._emitted_frames.keys()):
            if int(tick) >= int(keep_from):
                continue
            self._emitted_frames.pop(int(tick), None)


__all__ = ["RollbackController", "RollbackFrame"]



from __future__ import annotations

import datetime as dt
import logging
import os
import sys
from pathlib import Path

import structlog

_LEVELS: dict[str, int] = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}


def resolve_log_level(value: str | int) -> int:
    if isinstance(value, int):
        return int(value)
    level_name = str(value).strip().lower()
    resolved = _LEVELS.get(level_name)
    if resolved is None:
        supported = ", ".join(sorted(_LEVELS))
        raise ValueError(f"unsupported log level {value!r}; expected one of: {supported}")
    return int(resolved)


def default_component_log_path(*, base_dir: Path, component: str) -> Path:
    component_name = str(component).strip().lower() or "app"
    timestamp = dt.datetime.now(dt.timezone.utc).strftime("%Y%m%dT%H%M%S.%fZ")
    return base_dir / "logs" / component_name / f"{component_name}-pid{os.getpid()}-{timestamp}.log"


def _configure_structlog(*, level: int) -> list[structlog.typing.Processor]:
    processors: list[structlog.typing.Processor] = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.TimeStamper(fmt="iso", utc=True),
    ]
    structlog.configure(
        processors=[
            structlog.contextvars.merge_contextvars,
            *processors,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.make_filtering_bound_logger(int(level)),
        cache_logger_on_first_use=True,
    )
    return processors


def ensure_structlog_stdlib_defaults() -> None:
    config = structlog.get_config()
    logger_factory = config.get("logger_factory")
    if isinstance(logger_factory, structlog.stdlib.LoggerFactory):
        return
    _configure_structlog(level=logging.INFO)


def configure_component_logging(
    *,
    logger_name: str,
    component: str,
    log_file: Path,
    level: str | int = "info",
) -> Path:
    level_no = resolve_log_level(level)
    processors = _configure_structlog(level=int(level_no))

    resolved_log_file = Path(log_file).expanduser()
    resolved_log_file.parent.mkdir(parents=True, exist_ok=True)

    console_formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=list(processors),
        processor=structlog.dev.ConsoleRenderer(colors=bool(getattr(sys.stdout, "isatty", lambda: False)())),
    )
    file_formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=list(processors),
        processor=structlog.processors.JSONRenderer(sort_keys=True),
    )

    logger = logging.getLogger(str(logger_name))
    logger.handlers.clear()
    logger.propagate = False
    logger.setLevel(int(level_no))

    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setLevel(int(level_no))
    stream_handler.setFormatter(console_formatter)

    file_handler = logging.FileHandler(resolved_log_file, mode="a", encoding="utf-8")
    file_handler.setLevel(int(level_no))
    file_handler.setFormatter(file_formatter)

    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)

    structlog.stdlib.get_logger(str(logger_name)).info(
        "logging_configured",
        component=str(component),
        logger_name=str(logger_name),
        log_file=str(resolved_log_file),
        level=logging.getLevelName(level_no).lower(),
    )
    return resolved_log_file


__all__ = [
    "configure_component_logging",
    "default_component_log_path",
    "ensure_structlog_stdlib_defaults",
    "resolve_log_level",
]



from __future__ import annotations

import random

from grim.geom import Vec2

from ..creatures.spawn import SpawnId
from ..perks import PerkId
from ..weapons import WeaponId
from .helpers import (
    center_point,
    edge_midpoints,
    line_points,
    radial_points,
    random_angle,
    ring_points,
    spawn,
    spawn_at,
)
from .registry import register_quest
from .types import QuestContext, SpawnEntry


@register_quest(
    level="3.1",
    title="The Blighting",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_perk_id=PerkId.TOXIC_AVENGER,
    builder_address=0x00438050,
)
def build_3_1_the_blighting(ctx: QuestContext) -> list[SpawnEntry]:
    edges = edge_midpoints(ctx.width)
    edges_wide = edge_midpoints(ctx.width, offset=128.0)
    entries = [
        spawn_at(
            edges_wide.right,
            heading=0.0,
            spawn_id=SpawnId.ALIEN_CONST_RED_FAST_2B,
            trigger_ms=1500,
            count=2,
        ),
        spawn_at(edges_wide.left, heading=0.0, spawn_id=SpawnId.ALIEN_CONST_RED_FAST_2B, trigger_ms=1500, count=2),
        spawn(
            Vec2(896.0, 128.0), heading=0.0, spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_FAST_07, trigger_ms=2000, count=1,
        ),
        spawn(
            Vec2(128.0, 128.0), heading=0.0, spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_FAST_07, trigger_ms=2000, count=1,
        ),
        spawn(
            Vec2(128.0, 896.0), heading=0.0, spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_FAST_07, trigger_ms=2000, count=1,
        ),
        spawn(
            Vec2(896.0, 896.0), heading=0.0, spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_FAST_07, trigger_ms=2000, count=1,
        ),
    ]

    trigger = 4000
    for wave in range(8):
        if wave in (2, 4):
            entries.append(
                spawn_at(
                    edges_wide.left,
                    heading=0.0,
                    spawn_id=SpawnId.ALIEN_CONST_RED_FAST_2B,
                    trigger_ms=trigger,
                    count=4,
                ),
            )
        if wave in (3, 5):
            entries.append(
                spawn_at(
                    edges_wide.right,
                    heading=0.0,
                    spawn_id=SpawnId.ALIEN_CONST_RED_FAST_2B,
                    trigger_ms=trigger,
                    count=4,
                ),
            )
        spawn_id = SpawnId.AI1_ALIEN_BLUE_TINT_1A if wave % 2 == 0 else SpawnId.AI1_LIZARD_BLUE_TINT_1C
        edge = wave % 5
        if edge == 0:
            entries.append(
                spawn_at(
                    edges.right,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
            trigger += 15000
        elif edge == 1:
            entries.append(
                spawn_at(
                    edges.left,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
            trigger += 15000
        elif edge == 2:
            entries.append(
                spawn_at(
                    edges.bottom,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
            trigger += 15000
        elif edge == 3:
            entries.append(
                spawn_at(
                    edges.top,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
            trigger += 15000
        trigger += 1000
    return entries


@register_quest(
    level="3.2",
    title="Lizard Kings",
    time_limit_ms=180000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_weapon_id=WeaponId.MULTI_PLASMA,
    builder_address=0x00437710,
)
def build_3_2_lizard_kings(ctx: QuestContext) -> list[SpawnEntry]:
    center = center_point(ctx.width, ctx.height)
    entries = [
        spawn(
            Vec2(1152.0, 512.0),
            heading=0.0,
            spawn_id=SpawnId.FORMATION_CHAIN_LIZARD_4_11,
            trigger_ms=1500,
            count=1,
        ),
        spawn(
            Vec2(-128.0, 512.0),
            heading=0.0,
            spawn_id=SpawnId.FORMATION_CHAIN_LIZARD_4_11,
            trigger_ms=1500,
            count=1,
        ),
        spawn(
            Vec2(1152.0, 896.0),
            heading=0.0,
            spawn_id=SpawnId.FORMATION_CHAIN_LIZARD_4_11,
            trigger_ms=1500,
            count=1,
        ),
    ]
    trigger = 1500
    for pos, angle in ring_points(center, 256.0, 28, step=0.34906587):
        entries.append(
            spawn(
                pos,
                heading=-angle,
                spawn_id=SpawnId.LIZARD_RANDOM_31,
                trigger_ms=trigger,
                count=1,
            ),
        )
        trigger += 900
    return entries


@register_quest(
    level="3.3",
    title="The Killing",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_perk_id=PerkId.REGENERATION,
    builder_address=0x004384A0,
)
def build_3_3_the_killing(ctx: QuestContext, rng: random.Random | None = None) -> list[SpawnEntry]:
    rng = rng or random.Random()
    edges = edge_midpoints(ctx.width)
    entries: list[SpawnEntry] = []
    trigger = 2000
    for wave in range(10):
        rng.randrange(0x8000)
        rng.randrange(0x8000)
        spawn_cycle = wave % 3
        if spawn_cycle == 0:
            spawn_id = SpawnId.AI1_ALIEN_BLUE_TINT_1A
        elif spawn_cycle == 1:
            spawn_id = SpawnId.AI1_SPIDER_SP1_BLUE_TINT_1B
        else:
            spawn_id = SpawnId.AI1_LIZARD_BLUE_TINT_1C

        edge = wave % 5
        if edge == 0:
            entries.append(
                spawn_at(
                    edges.right,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
        elif edge == 1:
            entries.append(
                spawn_at(
                    edges.left,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
        elif edge == 2:
            entries.append(
                spawn_at(
                    edges.bottom,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
        elif edge == 3:
            entries.append(
                spawn_at(
                    edges.top,
                    heading=0.0,
                    spawn_id=spawn_id,
                    trigger_ms=trigger,
                    count=12,
                ),
            )
        else:
            for offset in (0, 1000, 2000):
                x = rng.randrange(0x300) + 0x80
                y = rng.randrange(0x300) + 0x80
                entries.append(
                    spawn(
                        Vec2(float(x), float(y)),
                        heading=0.0,
                        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_FAST_07,
                        trigger_ms=trigger + offset,
                        count=3,
                    ),
                )

        trigger += 6000
    return entries


@register_quest(
    level="3.4",
    title="Hidden Evil",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_weapon_id=WeaponId.SEEKER_ROCKETS,
    builder_address=0x00435A30,
)
def build_3_4_hidden_evil(ctx: QuestContext) -> list[SpawnEntry]:
    edges = edge_midpoints(ctx.width, ctx.height)
    return [
        spawn_at(edges.bottom, heading=0.0, spawn_id=SpawnId.ALIEN_CONST_PURPLE_GHOST_21, trigger_ms=500, count=50),
        spawn_at(edges.bottom, heading=0.0, spawn_id=SpawnId.ALIEN_CONST_GREEN_GHOST_22, trigger_ms=15000, count=30),
        spawn_at(
            edges.bottom, heading=0.0, spawn_id=SpawnId.ALIEN_CONST_GREEN_GHOST_SMALL_23, trigger_ms=25000, count=20,
        ),
        spawn_at(
            edges.bottom, heading=0.0, spawn_id=SpawnId.ALIEN_CONST_GREEN_GHOST_SMALL_23, trigger_ms=30000, count=30,
        ),
        spawn_at(edges.bottom, heading=0.0, spawn_id=SpawnId.ALIEN_CONST_GREEN_GHOST_22, trigger_ms=35000, count=30),
    ]


@register_quest(
    level="3.5",
    title="Surrounded By Reptiles",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_perk_id=PerkId.PYROMANIAC,
    builder_address=0x00438940,
)
def build_3_5_surrounded_by_reptiles(ctx: QuestContext) -> list[SpawnEntry]:
    entries: list[SpawnEntry] = []
    trigger = 1000
    for pos in line_points(Vec2(256.0, 256.0), Vec2(0.0, 102.4), 5):
        entries.append(
            spawn(
                Vec2(256.0, pos.y),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_SLOW_0D,
                trigger_ms=trigger,
                count=1,
            ),
        )
        entries.append(
            spawn(
                Vec2(768.0, pos.y),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_SLOW_0D,
                trigger_ms=trigger,
                count=1,
            ),
        )
        trigger += 800

    trigger = 8000
    for pos in line_points(Vec2(256.0, 256.0), Vec2(102.4, 0.0), 5):
        entries.append(
            spawn(
                Vec2(pos.x, 256.0),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_SLOW_0D,
                trigger_ms=trigger,
                count=1,
            ),
        )
        entries.append(
            spawn(
                Vec2(pos.x, 768.0),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_SLOW_0D,
                trigger_ms=trigger,
                count=1,
            ),
        )
        trigger += 800
    return entries


@register_quest(
    level="3.6",
    title="The Lizquidation",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_weapon_id=WeaponId.BLOW_TORCH,
    builder_address=0x00437C70,
)
def build_3_6_the_lizquidation(ctx: QuestContext) -> list[SpawnEntry]:
    entries: list[SpawnEntry] = []
    edges = edge_midpoints(ctx.width)
    trigger = 1500
    for wave in range(10):
        count = wave + 6
        entries.append(
            spawn_at(
                edges.right,
                heading=0.0,
                spawn_id=SpawnId.LIZARD_RANDOM_2E,
                trigger_ms=trigger,
                count=count,
            ),
        )
        entries.append(
            spawn_at(
                edges.left,
                heading=0.0,
                spawn_id=SpawnId.LIZARD_RANDOM_2E,
                trigger_ms=trigger,
                count=count,
            ),
        )
        if wave == 4:
            entries.append(
                spawn(
                    Vec2(ctx.width + 128.0, edges.right.y),
                    heading=0.0,
                    spawn_id=SpawnId.ALIEN_CONST_RED_FAST_2B,
                    trigger_ms=1500,
                    count=2,
                ),
            )
        trigger += 8000
    return entries


@register_quest(
    level="3.7",
    title="Spiders Inc.",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PLASMA_MINIGUN,
    unlock_perk_id=PerkId.NINJA,
    builder_address=0x004390D0,
)
def build_3_7_spiders_inc(ctx: QuestContext) -> list[SpawnEntry]:
    edges = edge_midpoints(ctx.width)
    center = center_point(ctx.width, ctx.height)
    entries = [
        spawn_at(edges.bottom, heading=0.0, spawn_id=SpawnId.SPIDER_SP1_AI7_TIMER_38, trigger_ms=500, count=1),
        spawn(
            Vec2(center.x + 64.0, edges.bottom.y),
            heading=0.0,
            spawn_id=SpawnId.SPIDER_SP1_AI7_TIMER_38,
            trigger_ms=500,
            count=1,
        ),
        spawn_at(edges.top, heading=0.0, spawn_id=SpawnId.SPIDER_SP1_CONST_BLUE_40, trigger_ms=500, count=4),
    ]

    trigger = 17000
    step_count = 0
    while trigger < 107000:
        count = step_count // 2 + 3
        entries.append(
            spawn_at(
                edges.bottom,
                heading=0.0,
                spawn_id=SpawnId.SPIDER_SP1_AI7_TIMER_38,
                trigger_ms=trigger,
                count=count,
            ),
        )
        entries.append(
            spawn_at(
                edges.top,
                heading=0.0,
                spawn_id=SpawnId.SPIDER_SP1_AI7_TIMER_38,
                trigger_ms=trigger,
                count=count,
            ),
        )
        trigger += 6000
        step_count += 1
    return entries


@register_quest(
    level="3.8",
    title="Lizard Raze",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_weapon_id=WeaponId.ROCKET_MINIGUN,
    builder_address=0x00438840,
)
def build_3_8_lizard_raze(ctx: QuestContext) -> list[SpawnEntry]:
    entries: list[SpawnEntry] = []
    edges = edge_midpoints(ctx.width)
    trigger = 1500
    while trigger < 91500:
        entries.append(
            spawn_at(
                edges.right,
                heading=0.0,
                spawn_id=SpawnId.LIZARD_RANDOM_2E,
                trigger_ms=trigger,
                count=6,
            ),
        )
        entries.append(
            spawn_at(
                edges.left,
                heading=0.0,
                spawn_id=SpawnId.LIZARD_RANDOM_2E,
                trigger_ms=trigger,
                count=6,
            ),
        )
        trigger += 6000
    entries.extend(
        [
            spawn(
                Vec2(128.0, 256.0),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_FAST_0C,
                trigger_ms=10000,
                count=1,
            ),
            spawn(
                Vec2(128.0, 384.0),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_FAST_0C,
                trigger_ms=10000,
                count=1,
            ),
            spawn(
                Vec2(128.0, 512.0),
                heading=0.0,
                spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_FAST_0C,
                trigger_ms=10000,
                count=1,
            ),
        ],
    )
    return entries


@register_quest(
    level="3.9",
    title="Deja vu",
    time_limit_ms=120000,
    start_weapon_id=WeaponId.GAUSS_GUN,
    unlock_perk_id=PerkId.HIGHLANDER,
    builder_address=0x00437920,
)
def build_3_9_deja_vu(ctx: QuestContext, rng: random.Random | None = None) -> list[SpawnEntry]:
    rng = rng or random.Random()
    entries: list[SpawnEntry] = []
    center = center_point(ctx.width, ctx.height)
    trigger = 2000
    step = 2000
    while step > 560:
        angle = random_angle(rng)
        for pos in radial_points(center, angle, 0x54, 0xFC, 0x2A):
            entries.append(
                spawn(
                    pos,
                    heading=0.0,
                    spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_SLOW_0D,
                    trigger_ms=trigger,
                    count=1,
                ),
            )
        trigger += step
        step -= 0x50
    return entries


@register_quest(
    level="3.10",
    title="Zombie Masters",
    time_limit_ms=300000,
    start_weapon_id=WeaponId.PISTOL,
    unlock_weapon_id=WeaponId.JACKHAMMER,
    builder_address=0x004360A0,
)
def build_3_10_zombie_masters(ctx: QuestContext) -> list[SpawnEntry]:
    return [
        spawn(
            Vec2(256.0, 256.0),
            heading=0.0,
            spawn_id=SpawnId.ZOMBIE_BOSS_SPAWNER_00,
            trigger_ms=1000,
            count=ctx.player_count,
        ),
        spawn(Vec2(512.0, 256.0), heading=0.0, spawn_id=SpawnId.ZOMBIE_BOSS_SPAWNER_00, trigger_ms=6000, count=1),
        spawn(
            Vec2(768.0, 256.0),
            heading=0.0,
            spawn_id=SpawnId.ZOMBIE_BOSS_SPAWNER_00,
            trigger_ms=14000,
            count=ctx.player_count,
        ),
        spawn(
            Vec2(768.0, 768.0),
            heading=0.0,
            spawn_id=SpawnId.ZOMBIE_BOSS_SPAWNER_00,
            trigger_ms=18000,
            count=1,
        ),
    ]


__all__ = [
    "QuestContext",
    "SpawnEntry",
    "build_3_1_the_blighting",
    "build_3_2_lizard_kings",
    "build_3_3_the_killing",
    "build_3_4_hidden_evil",
    "build_3_5_surrounded_by_reptiles",
    "build_3_6_the_lizquidation",
    "build_3_7_spiders_inc",
    "build_3_8_lizard_raze",
    "build_3_9_deja_vu",
    "build_3_10_zombie_masters",
]



from __future__ import annotations

from ...weapon_runtime.assign import weapon_assign_player
from ...weapon_runtime.availability import weapon_pick_random_available
from ...weapons import WeaponId
from ..ids import PerkId
from ..runtime.apply_context import PerkApplyCtx
from ..runtime.hook_types import PerkHooks


def apply_random_weapon(ctx: PerkApplyCtx) -> None:
    current = int(ctx.owner.weapon_id)
    weapon_id = int(current)
    for _ in range(100):
        candidate = int(weapon_pick_random_available(ctx.state))
        weapon_id = candidate
        if candidate != int(WeaponId.PISTOL) and candidate != current:
            break
    weapon_assign_player(ctx.owner, weapon_id, state=ctx.state)


HOOKS = PerkHooks(
    perk_id=PerkId.RANDOM_WEAPON,
    apply_handler=apply_random_weapon,
)



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field

from ..replay.types import PackedPlayerInput
from .protocol import (
    INPUT_DELAY_TICKS,
    INPUT_STALL_TIMEOUT_MS,
    STATE_HASH_PERIOD_TICKS,
    InputBatch,
    InputSample,
    PauseState,
    TickFrame,
)

CLIENT_MAX_CAPTURE_LEAD_TICKS = 1
CLIENT_MAX_RESEND_SAMPLES = 64
CLIENT_MAX_SENT_HISTORY_TICKS = 256


@dataclass(slots=True)
class HostLockstepState:
    player_count: int
    input_delay_ticks: int = INPUT_DELAY_TICKS
    input_stall_timeout_ms: int = INPUT_STALL_TIMEOUT_MS
    state_hash_period_ticks: int = STATE_HASH_PERIOD_TICKS
    _inputs_by_tick: dict[int, dict[int, PackedPlayerInput]] = field(default_factory=dict)
    _next_emit_tick: int = 0
    _last_progress_ms: int = 0
    _paused: bool = False

    @property
    def next_emit_tick(self) -> int:
        return int(self._next_emit_tick)

    @property
    def buffered_tick_count(self) -> int:
        return int(len(self._inputs_by_tick))

    @property
    def paused(self) -> bool:
        return bool(self._paused)

    @property
    def last_progress_ms(self) -> int:
        return int(self._last_progress_ms)

    def waiting_for_inputs(self, *, tick_index: int | None = None) -> int:
        if tick_index is None:
            tick_index = int(self._next_emit_tick)
        tick_inputs = self._inputs_by_tick.get(int(tick_index), {})
        return max(0, int(self.player_count) - int(len(tick_inputs)))

    def submit_input_sample(self, *, slot_index: int, tick_index: int, packed_input: PackedPlayerInput) -> None:
        if int(slot_index) < 0 or int(slot_index) >= int(self.player_count):
            return
        if int(tick_index) < int(self._next_emit_tick):
            return
        tick_inputs = self._inputs_by_tick.setdefault(int(tick_index), {})
        tick_inputs[int(slot_index)] = list(packed_input)

    def submit_input_batch(self, batch: InputBatch) -> None:
        slot = int(batch.slot_index)
        for sample in batch.samples:
            self.submit_input_sample(
                slot_index=int(slot),
                tick_index=int(sample.tick_index),
                packed_input=sample.packed_input,
            )

    def _tick_complete(self, tick_index: int) -> bool:
        tick_inputs = self._inputs_by_tick.get(int(tick_index))
        if tick_inputs is None:
            return False
        if len(tick_inputs) < int(self.player_count):
            return False
        return all(int(slot) in tick_inputs for slot in range(int(self.player_count)))

    def pop_ready_frames(
        self,
        *,
        now_ms: int,
        command_hash_by_tick: dict[int, str] | None = None,
        state_hash_by_tick: dict[int, str] | None = None,
    ) -> list[TickFrame]:
        frames: list[TickFrame] = []
        while self._tick_complete(int(self._next_emit_tick)):
            tick = int(self._next_emit_tick)
            tick_inputs = self._inputs_by_tick.pop(tick, {})
            ordered_inputs = [list(tick_inputs[slot]) for slot in range(int(self.player_count))]
            command_hash = ""
            if command_hash_by_tick is not None:
                command_hash = str(command_hash_by_tick.get(int(tick), ""))
            state_hash = ""
            if state_hash_by_tick is not None and (int(tick) % int(self.state_hash_period_ticks)) == 0:
                state_hash = str(state_hash_by_tick.get(int(tick), ""))
            frames.append(
                TickFrame(
                    tick_index=int(tick),
                    frame_inputs=ordered_inputs,
                    command_hash=str(command_hash),
                    state_hash=str(state_hash),
                ),
            )
            self._next_emit_tick += 1
            self._last_progress_ms = int(now_ms)

        return frames

    def update_pause_state(self, *, now_ms: int) -> PauseState | None:
        tick_inputs = self._inputs_by_tick.get(int(self._next_emit_tick), {})
        waiting_for = max(0, int(self.player_count) - len(tick_inputs))
        should_pause = waiting_for > 0 and (int(now_ms) - int(self._last_progress_ms)) >= int(self.input_stall_timeout_ms)
        if bool(should_pause) == bool(self._paused):
            return None
        self._paused = bool(should_pause)
        if should_pause:
            return PauseState(paused=True, reason="waiting_input")
        return PauseState(paused=False, reason="")


@dataclass(slots=True)
class ClientLockstepState:
    local_slot_index: int
    input_delay_ticks: int = INPUT_DELAY_TICKS
    input_stall_timeout_ms: int = INPUT_STALL_TIMEOUT_MS
    max_resend_samples: int = CLIENT_MAX_RESEND_SAMPLES
    max_sent_history_ticks: int = CLIENT_MAX_SENT_HISTORY_TICKS
    _capture_tick: int = 0
    _sent_inputs: dict[int, PackedPlayerInput] = field(default_factory=dict)
    _canonical_by_tick: dict[int, TickFrame] = field(default_factory=dict)
    _next_consume_tick: int = 0
    _last_progress_ms: int = 0
    _paused: bool = False
    _pending_desync: deque[tuple[int, str, str]] = field(default_factory=deque)

    @property
    def next_consume_tick(self) -> int:
        return int(self._next_consume_tick)

    @property
    def capture_tick(self) -> int:
        return int(self._capture_tick)

    @property
    def buffered_frame_count(self) -> int:
        return int(len(self._canonical_by_tick))

    @property
    def paused(self) -> bool:
        return bool(self._paused)

    @property
    def last_progress_ms(self) -> int:
        return int(self._last_progress_ms)

    def queue_local_input(self, packed_input: PackedPlayerInput) -> InputBatch:
        # Keep the joiner's "capture clock" close to lockstep progress to avoid
        # scheduling inputs far in the future (which manifests as extra input lag).
        max_capture_tick = int(self._next_consume_tick) + int(CLIENT_MAX_CAPTURE_LEAD_TICKS)
        if int(self._capture_tick) > int(max_capture_tick):
            self._capture_tick = int(max_capture_tick)

        target_tick = int(self._capture_tick + int(self.input_delay_ticks))
        self._sent_inputs[int(target_tick)] = list(packed_input)

        samples: list[InputSample] = []
        oldest_tick = max(0, int(self._next_consume_tick))
        history_ticks = int(self.max_sent_history_ticks)
        if history_ticks > 0:
            oldest_tick = max(int(oldest_tick), int(target_tick) - int(history_ticks) + 1)
        max_samples = max(1, int(self.max_resend_samples))
        sample_count = 0
        for tick in range(int(target_tick), int(oldest_tick) - 1, -1):
            value = self._sent_inputs.get(int(tick))
            if value is None:
                continue
            samples.append(InputSample(tick_index=int(tick), packed_input=list(value)))
            sample_count += 1
            if int(sample_count) >= int(max_samples):
                break

        # Keep memory bounded while preserving any still-unacknowledged ticks.
        min_keep_tick = max(0, int(self._next_consume_tick))
        if history_ticks > 0:
            min_keep_tick = max(int(min_keep_tick), int(target_tick) - int(history_ticks) + 1)
        max_keep_tick = int(target_tick)
        for tick in list(self._sent_inputs):
            if int(tick) < int(min_keep_tick) or int(tick) > int(max_keep_tick):
                self._sent_inputs.pop(int(tick), None)

        self._capture_tick += 1
        return InputBatch(slot_index=int(self.local_slot_index), samples=samples)

    def ingest_tick_frame(self, frame: TickFrame, *, now_ms: int, local_command_hash: str = "") -> None:
        tick = int(frame.tick_index)
        self._canonical_by_tick[int(tick)] = frame
        self._last_progress_ms = int(now_ms)
        remote_hash = str(frame.command_hash or "")
        if remote_hash and str(local_command_hash or "") and str(local_command_hash) != str(remote_hash):
            self._pending_desync.append((int(tick), str(remote_hash), str(local_command_hash)))

    def pop_canonical_frame(self) -> TickFrame | None:
        tick = int(self._next_consume_tick)
        frame = self._canonical_by_tick.pop(int(tick), None)
        if frame is None:
            return None
        self._next_consume_tick += 1
        return frame

    def pop_desync_notice(self) -> tuple[int, str, str] | None:
        if not self._pending_desync:
            return None
        return self._pending_desync.popleft()

    def update_pause_state(self, *, now_ms: int) -> PauseState | None:
        frame_ready = int(self._next_consume_tick) in self._canonical_by_tick
        should_pause = (not frame_ready) and (
            int(now_ms) - int(self._last_progress_ms)
        ) >= int(self.input_stall_timeout_ms)
        if bool(should_pause) == bool(self._paused):
            return None
        self._paused = bool(should_pause)
        if should_pause:
            return PauseState(paused=True, reason="waiting_tick_frame")
        return PauseState(paused=False, reason="")
